// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"errors"
	"os"

	"github.com/spf13/cobra"
	xbus "xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/api"
)

// actorAcceptCmd represents the actor accept sub-command.
var actorAcceptCmd = &cobra.Command{
	Use:   "accept",
	Short: "Accept an actor.",
	Long: "Accept a pending actor so it can be allowed to connect to Xbus. An actor ID or name " +
		"is required.",
	Run: acceptActor,
}

// acceptActor parses arguments to accept a pending actor.
func acceptActor(cmd *cobra.Command, args []string) {
	var identifier string
	v := cmd.Flag("id")
	switch {
	case v.Changed:
		identifier = v.Value.String()
	case len(args) == 1:
		identifier = args[0]
	default:
		usage(cmd)
	}
	os.Exit(doAcceptActor(identifier))
}

// doAcceptActor accepts a pending actor request.
func doAcceptActor(identifier string) int {
	conn, err := GetClient(false)
	if err != nil {
		Logger.Error().Msgf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	actor, err := conn.ActorFind(identifier)
	if err != nil {
		if errors.Is(err, xbus.ErrNotFound) {
			Logger.Error().Msgf("actor not found")
		} else {
			Logger.Error().Msgf("could not find actor: %s", err)
		}
		return -1
	}

	// TODO Find the actor based on the name.

	// Send an "accept" request.
	repActor, err := conn.ActorAccept(actor.GetIDAsUUID())
	if err != nil {
		Logger.Error().Msgf("Could not accept the actor %q: %s", actor.ID, err)
		return 1
	}
	if repActor.Status == api.Actor_PENDING {
		Logger.Error().Msgf("Could not accept the actor %q: Still pending.", actor.ID)
		return 1
	}
	Logger.Info().Msgf("Successfully accepted the actor %q (ID %q).", repActor.Name, actor.ID)
	return 0
}

func init() {
	actorCmd.AddCommand(actorAcceptCmd)
}
