// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"errors"
	"os"
	"strings"

	"github.com/spf13/cobra"
	xbus "xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/api/control"
)

var actorConfigSetCmd = &cobra.Command{
	Use:   "set [ actor ] <key>=<value>",
	Short: "Set an actor configuration entry",
	Long:  "Set an actor configuration entry",
	Run: func(cmd *cobra.Command, args []string) {
		// Default action, display the config entries

		if len(args) < 1 || len(args) > 2 {
			usage(cmd)
		}

		conn, err := GetClient(false)
		if err != nil {
			Logger.Error().Msgf("Cannot connect to the NATS server: %s", err)
			os.Exit(1)
		}
		var identifier string
		if v := cmd.Flag("id"); v.Changed {
			identifier = v.Value.String()
		} else if v := cmd.Flag("name"); v.Changed {
			identifier = v.Value.String()
		} else if len(args) == 2 {
			identifier = args[0]
			args = args[1:]
		} else {
			usage(cmd)
		}

		actor, err := conn.ActorFind(identifier)
		if err != nil {
			if errors.Is(err, xbus.ErrNotFound) {
				Logger.Error().Msgf("actor not found")
			} else {
				Logger.Error().Msgf("could not find actor: %s", err)
			}
			os.Exit(1)
		}

		splitted := strings.SplitN(args[0], "=", 2)
		if len(splitted) != 2 {
			usage(cmd)
		}
		key := splitted[0]
		value := splitted[1]

		found := false
		for i, entry := range actor.Config {
			if entry.Key == key {
				actor.Config[i].Value = value
				found = true
			}
		}
		if !found {
			actor.Config = append(actor.Config, &api.ConfigEntry{
				Key:   key,
				Value: value,
			})
		}

		if _, err := conn.Control.Actor.Update(&control.ActorRequest{
			ActorList: []*api.Actor{actor},
		}); err != nil {
			Logger.Error().Msgf("Error updating: %s", err)
		}
	},
}

func init() {
	actorConfigCmd.AddCommand(actorConfigSetCmd)
}
