// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"bufio"
	"context"
	"fmt"
	"os"
	"os/signal"
	"regexp"
	"strconv"
	"strings"
	"time"

	"github.com/olekukonko/tablewriter"
	"github.com/spf13/cobra"
	"google.golang.org/protobuf/types/known/timestamppb"
	xbus "xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/api/control"
)

var durationRe = regexp.MustCompile("^([0-9]+)([smhdw])$")
var unixEpochRe = regexp.MustCompile("^[0-9]+$")

func readDatetimeArg(arg string) (time.Time, error) {
	if matches := durationRe.FindStringSubmatch(arg); len(matches) != 0 {
		// its a duration, lets interpret it
		qty, err := strconv.Atoi(matches[1])
		if err != nil {
			return time.Time{}, err
		}
		var duration time.Duration
		switch matches[2] {
		case "s":
			duration = time.Second * time.Duration(qty)
		case "m":
			duration = time.Minute * time.Duration(qty)
		case "h":
			duration = time.Hour * time.Duration(qty)
		case "d":
			duration = time.Hour * 24 * time.Duration(qty)
		case "w":
			duration = time.Hour * 24 * 7 * time.Duration(qty)
		default:
			return time.Time{}, fmt.Errorf("Unknown duration unit: %s", matches[2])
		}
		return time.Now().Add(-duration), nil
	}
	if unixEpochRe.MatchString(arg) {
		// its a unix timestamp
		qty, err := strconv.Atoi(arg)
		if err != nil {
			return time.Time{}, err
		}
		return time.Unix(int64(qty), 0), nil
	}
	// It's not a duration nor a unix timestamp, attempt to parse it as a
	// datetime, allowing various iso-like formats
	for _, format := range []string{
		time.RFC3339,
		"2006-01-02T15:04:05Z0700",
		"2006-01-02T15:04:05",
		"2006-01-02 15:04:05Z07:00",
		"2006-01-02 15:04:05Z0700",
		"2006-01-02 15:04:05",
		"2006-01-02",
	} {
		t, err := time.ParseInLocation(format, arg, time.Local)
		if err == nil {
			return t, nil
		}
	}
	return time.Time{}, fmt.Errorf("Could not parse as a valid duration or date: %s", arg)
}

func readUUIDList(s string) (api.UUIDSlice, error) {
	if s == "" {
		return nil, nil
	}
	var sep string
	if strings.Contains(s, " ") {
		sep = " "
	} else {
		sep = ","
	}

	var (
		splitted = strings.Split(s, sep)
		list     = make([]api.UUID, len(splitted))
	)

	for i, sid := range splitted {
		id, err := api.UUIDFromString(sid)
		if err != nil {
			return nil, err
		}
		list[i] = id
	}

	return list, nil
}

func contextCancelOnSignal(ctx context.Context, s ...os.Signal) (context.Context, context.CancelFunc) {
	ctx, cancel := context.WithCancel(ctx)

	c := make(chan os.Signal, 1)
	for _, sig := range s {
		signal.Notify(c, sig)
	}

	go func() {
		select {
		case <-ctx.Done():
		case <-c:
			signal.Stop(c)
			cancel()
		}
	}()

	return ctx, cancel
}

var psCmd = &cobra.Command{
	Use:   "ps",
	Short: "Process list",
	Run: func(cmd *cobra.Command, args []string) {
		c, err := GetClient(false)
		if err != nil {
			return
		}
		filter, err := getProcessFilter(
			c, args, filterFlags{IncludeLiveStatuses: true})
		if err != nil {
			Logger.Error().Msgf("Invalid filter: %s", err)
			return
		}
		processList, err := getProcessList(c, filter, "", false)
		if err != nil {
			Logger.Error().Msgf("Error: %s", err)
			return
		}
		pipelineList, err := c.PipelineQuery("")
		if err != nil {
			Logger.Error().Msgf("%s", err)
			return
		}
		actorList, err := c.ActorList(nil)
		if err != nil {
			Logger.Error().Msgf("%s", err)
			return
		}
		printProcessList(processList, pipelineList, actorList)
	},
}

func doCommand(
	args []string, defaultFilterFlags filterFlags,
	command control.ProcessControlRequest_Command, confirm bool,
) int {
	c, err := GetClient(false)
	if err != nil {
		return 1
	}
	filter, err := getProcessFilter(
		c, args, defaultFilterFlags)
	if err != nil {
		Logger.Error().Msgf("Invalid filter: %s", err)
		return 1
	}
	msg := ""
	if confirm {
		msg = fmt.Sprintf(
			"Will run the command %s on %%d process(es). Do you confirm (yes/NO)?", command,
		)
	}
	list, err := getProcessList(c, filter, msg, false)
	if err != nil {
		fmt.Println(err)
		os.Exit(1)
	}
	if len(list) == 0 {
		fmt.Println("No process matched the filter")
		os.Exit(0)
	}
	errorCount := 0
	for i, p := range list {
		err = c.ProcessControl(p.GetIDAsUUID(), command)
		if err != nil {
			errorCount++
			Logger.Error().Msgf("Process %s: %s", p.ID, err)
		}
		fmt.Printf("\r%d success, %d errors / %d     ", (i + 1 - errorCount), errorCount, len(list))
	}
	fmt.Printf("\n%d success, %d errors", len(list)-errorCount, errorCount)
	if errorCount != 0 {
		Logger.Error().Msgf("Encountered %d errors, the command could not be applied to all processes.", errorCount)
		return 1
	}
	return 0
}

func getProcessList(client *xbus.Client, filter *api.ProcessFilter, confirmText string, quiet bool) ([]*api.Process, error) {
	var processList []*api.Process
	if !quiet {
		fmt.Printf("Loading process list...")
	}
	ctx, _ := contextCancelOnSignal(
		context.Background(), os.Interrupt,
	)
	if err := client.ProcessQuery(ctx, filter, func(list []*api.Process) {
		processList = append(processList, list...)
		if !quiet {
			fmt.Printf("\rLoading process list... %d", len(processList))
		}
	}); err != nil {
		return nil, err
	}
	if !quiet {
		fmt.Printf("\rLoaded %d processes       \n", len(processList))
	}
	if !quiet && len(processList) != 0 && confirmText != "" {
		printProcessList(processList, nil, nil)
		reader := bufio.NewReader(os.Stdin)
		fmt.Printf(confirmText, len(processList))
		text, _ := reader.ReadString('\n')
		if !(text == "Y\n" || text == "y\n" || text == "yes\n") {
			return nil, nil
		}
	}
	return processList, nil
}

func printProcessList(
	processList []*api.Process, pipelineList []*api.PipelineInfo, actorList []*api.Actor,
) {
	var pList = processList
	if len(processList) > 1000 {
		pList = pList[:1000]
	}

	pipelines := make(map[api.UUID]*api.PipelineInfo)
	for _, g := range pipelineList {
		pipelines[g.GetIDAsUUID()] = g
	}
	actors := make(map[api.UUID]*api.Actor)
	for _, a := range actorList {
		actors[a.GetIDAsUUID()] = a
	}
	table := tablewriter.NewWriter(os.Stdout)
	table.SetHeader([]string{"id", "created at", "pipeline", "emitter", "status"})

	for _, p := range pList {
		createdAt := ""
		if p.CreatedAt != nil {
			createdAt = p.CreatedAt.AsTime().Format(time.RFC3339)
		}
		pipeline := p.GetPipelineIDAsUUID().String()
		if g, ok := pipelines[p.GetPipelineIDAsUUID()]; ok {
			pipeline = g.GetName() + " " + g.GetVersion()
		}
		actor := p.GetTriggerEmitterIDAsUUID().String()
		if a, ok := actors[p.GetTriggerEmitterIDAsUUID()]; ok {
			actor = a.GetName()
		}
		table.Append([]string{
			p.GetIDAsUUID().String(),
			createdAt,
			pipeline,
			actor,
			p.Status.String(),
		})
	}
	table.Render()
	if len(processList) > 1000 {
		fmt.Printf("+ %d more processes...\n", len(processList)-1000)
	}
}

type filterFlags struct {
	GroupID                   string
	ReplayOf                  string
	CreatedAtFrom             string
	CreatedAtTo               string
	StatusChangedFrom         string
	StatusChangedTo           string
	IncludeStatusInitial      bool
	IncludeStatusRunning      bool
	IncludeStatusPaused       bool
	IncludeStatusDone         bool
	IncludeStatusError        bool
	IncludeLiveStatuses       bool
	IncludeTerminatedStatuses bool
	IncludeAllStatuses        bool
	AckedResponseOnly         bool
	NotAckedResponseOnly      bool
	EnvelopeIDs               string
	Pipelines                 string

	// start 3.0 compatibility
	StatusChangedSince   string
	StatusUnchangedSince string
	// end 3.0 compatibility
}

var globalFlags filterFlags

func getProcessFilter(client *xbus.Client, args []string, defaults filterFlags) (*api.ProcessFilter, error) {
	var (
		flags  = globalFlags
		filter api.ProcessFilter
	)

	if (globalFlags == filterFlags{}) && len(args) == 0 {
		flags = defaults
	}

	if ids, err := readUUIDList(flags.GroupID); err == nil {
		filter.GroupID = ids.ToBytes()
	} else {
		return nil, err
	}

	if ids, err := readUUIDList(flags.ReplayOf); err == nil {
		filter.ReplayOf = ids.ToBytes()
	} else {
		return nil, err
	}

	if flags.CreatedAtFrom != "" {
		t, err := readDatetimeArg(flags.CreatedAtFrom)
		if err != nil {
			return nil, fmt.Errorf("error reading 'created-at-from': %w", err)
		}
		filter.CreatedAt = &api.TimeRange{From: timestamppb.New(t)}
	}

	if flags.CreatedAtTo != "" {
		t, err := readDatetimeArg(flags.CreatedAtTo)
		if err != nil {
			return nil, fmt.Errorf("error reading 'created-at-to': %w", err)
		}
		if filter.CreatedAt == nil {
			filter.CreatedAt = &api.TimeRange{}
		}
		filter.CreatedAt.To = timestamppb.New(t)
	}

	// start 3.0 compatibility
	if flags.StatusChangedSince != "" {
		if flags.StatusChangedFrom != "" {
			return nil, fmt.Errorf("--status-changed-since and --status-changed-from conflicts. Use '--status-changed-from' only")
		}
		flags.StatusChangedFrom = flags.StatusChangedSince
		Logger.Info().Msgf("WARNING! --status-changed-since is deprecated and will be dropped in xbusctl 3.3. Please use --status-changed-since")
	}
	if flags.StatusUnchangedSince != "" {
		if flags.StatusChangedTo != "" {
			return nil, fmt.Errorf("--status-unchanged-since and --status-changed-to conflicts. Use '--status-changed-to' only")
		}
		flags.StatusChangedTo = flags.StatusUnchangedSince
		Logger.Info().Msgf("WARNING! --status-unchanged-since is deprecated and will be dropped in xbusctl 3.3. Please use --status-changed-to")
	}
	// end 3.0 compatibility

	if flags.StatusChangedFrom != "" {
		t, err := readDatetimeArg(flags.StatusChangedFrom)
		if err != nil {
			return nil, fmt.Errorf("error reading 'status-changed-from': %w", err)
		}
		filter.StatusChanged = &api.TimeRange{From: timestamppb.New(t)}

		// start 3.0 compatibility
		filter.StatusChangedFrom = timestamppb.New(t)
		// end 3.0 compatibility
	}

	if flags.StatusChangedTo != "" {
		t, err := readDatetimeArg(flags.StatusChangedTo)
		if err != nil {
			return nil, fmt.Errorf("error reading 'status-changed-to': %w", err)
		}
		if filter.StatusChanged == nil {
			filter.StatusChanged = &api.TimeRange{}
		}
		filter.StatusChanged.To = timestamppb.New(t)

		// start 3.0 compatibility
		filter.StatusChangedTo = timestamppb.New(t)
		// end 3.0 compatibility
	}

	if flags.IncludeAllStatuses || flags.IncludeLiveStatuses {
		flags.IncludeStatusInitial = true
		flags.IncludeStatusRunning = true
		flags.IncludeStatusPaused = true
	}

	if flags.IncludeAllStatuses || flags.IncludeTerminatedStatuses {
		flags.IncludeStatusDone = true
		flags.IncludeStatusError = true
	}

	if flags.IncludeStatusInitial {
		filter.Status = append(filter.Status, api.Process_INITIAL)
	}

	if flags.IncludeStatusRunning {
		filter.Status = append(filter.Status, api.Process_RUNNING)
	}

	if flags.IncludeStatusPaused {
		filter.Status = append(filter.Status, api.Process_PAUSED)
	}

	if flags.IncludeStatusDone {
		filter.Status = append(filter.Status, api.Process_DONE)
	}

	if flags.IncludeStatusError {
		filter.Status = append(filter.Status, api.Process_ERROR)
	}

	if flags.AckedResponseOnly {
		filter.ResultAcked = append(filter.ResultAcked, true)
	}

	if flags.NotAckedResponseOnly {
		filter.ResultAcked = append(filter.ResultAcked, false)
	}

	if ids, err := readUUIDList(flags.EnvelopeIDs); err == nil {
		filter.EnvelopeID = ids.ToBytes()
	} else {
		return nil, err
	}

	if flags.Pipelines != "" {
		allPipelines, err := client.PipelineQuery("")
		if err != nil {
			return nil, err
		}
		list := strings.Split(flags.Pipelines, ",")
		for _, item := range list {
			var (
				name  = item
				match = func(v string) bool { return true }
				found bool
			)
			for _, op := range []struct {
				op string
				fn func(v1, v2 string) bool
			}{
				{"==", func(v1, v2 string) bool {
					return v1 == v2
				}},
				{"=", func(v1, v2 string) bool {
					return v1 == v2
				}},
			} {
				if i := strings.Index(item, op.op); i != -1 {
					name = item[:i]
					match = func(v string) bool {
						return op.fn(v, item[i+len(op.op):])
					}
					break
				}
			}

			for _, pipeline := range allPipelines {
				if pipeline.GetName() == name && match(pipeline.GetVersion()) {
					found = true
					filter.PipelineID = append(filter.PipelineID, pipeline.ID)
				}
			}
			if !found {
				return nil, fmt.Errorf("No graph matching '%s'", item)
			}
		}
	}

	for _, arg := range args {
		id, err := api.UUIDFromString(arg)
		if err != nil {
			return nil, fmt.Errorf("invalid processid: %s (%w)", arg, err)
		}
		filter.ID = append(filter.ID, id.ToBytes())
	}

	return &filter, nil
}

func init() {
	RootCmd.AddCommand(psCmd)

	flags := psCmd.PersistentFlags()

	// A few flags to create filters
	flags.StringVar(
		&globalFlags.GroupID,
		"group-id", "",
		"Processes that belong to a group. Expects a comma-separated list of process group ids.",
	)
	flags.StringVar(
		&globalFlags.ReplayOf,
		"replay-of", "",
		"Processes that are a replay of the given processes. "+
			"Expects a comma-separated list of process group ids.",
	)
	flags.StringVar(
		&globalFlags.CreatedAtFrom,
		"created-at-from", "",
		"Process was created at or after this date (or duration before now)",
	)
	flags.StringVar(
		&globalFlags.CreatedAtTo,
		"created-at-to", "",
		"Process was created before or at this date (or duration before now)",
	)

	flags.StringVar(
		&globalFlags.StatusChangedFrom,
		"status-changed-from", "",
		"Status of the process changed at least once after this date (or duration before now)")

	flags.StringVar(
		&globalFlags.StatusChangedTo,
		"status-changed-to", "",
		"Status of the process did not change after this date (or duration before now)")

	flags.StringVar(
		&globalFlags.StatusChangedSince,
		"status-changed-since", "",
		"Deprecated alias for --status-changed-from")

	flags.StringVar(
		&globalFlags.StatusUnchangedSince,
		"status-unchanged-since", "",
		"Deprecated alias for --status-changed-to")

	flags.BoolVar(
		&globalFlags.IncludeStatusInitial,
		"status-initial", false,
		"Include processes in the 'initial' status")

	flags.BoolVar(
		&globalFlags.IncludeStatusRunning,
		"status-running", false,
		"Include processes in the 'running' status")

	flags.BoolVar(
		&globalFlags.IncludeStatusPaused,
		"status-paused", false,
		"Include processes in the 'paused' status")

	flags.BoolVar(
		&globalFlags.IncludeStatusDone,
		"status-done", false,
		"Include processes in the 'done' status")

	flags.BoolVar(
		&globalFlags.IncludeStatusError,
		"status-error", false,
		"Include processes in the 'error' status")

	flags.BoolVar(
		&globalFlags.IncludeLiveStatuses,
		"live", false,
		"Include processes that are not terminated")

	flags.BoolVar(
		&globalFlags.IncludeTerminatedStatuses,
		"terminated", false,
		"Include processes that are terminated")

	flags.BoolVar(
		&globalFlags.IncludeAllStatuses,
		"all", false,
		"Include processes with any status",
	)

	flags.BoolVar(
		&globalFlags.AckedResponseOnly,
		"result-acked", false,
		"Include processes for which the result is acked")

	flags.BoolVar(
		&globalFlags.NotAckedResponseOnly,
		"result-not-acked", false,
		"Include processes for which the result is not acked")

	flags.StringVar(
		&globalFlags.EnvelopeIDs,
		"envelope-id", "",
		"Include processes that are related to one on the given envelopes",
	)

	flags.StringVar(
		&globalFlags.Pipelines,
		"pipeline", "",
		"Include pipelines matching the selection (ex: name1,name2==version,name3)",
	)
}
