// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"bufio"
	"bytes"
	"context"
	"fmt"
	"io"
	"os"
	"strconv"
	"strings"
	"sync"

	"github.com/spf13/cobra"
	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/api/control"

	"xbus.io/xbus/utils"
)

var psExportCmd = &cobra.Command{
	Use:   "export",
	Short: "Export process data",
	Run: func(cmd *cobra.Command, args []string) {
		var (
			output            = cmd.Flag("output").Value.String()
			yaml              = cmd.Flag("yaml").Value.String() == "true"
			quiet             = cmd.Flag("quiet").Value.String() == "true"
			full              = cmd.Flag("full").Value.String() == "true"
			exportEnvelope    = cmd.Flag("export-envelope").Value.String() == "true"
			exportJobs        = cmd.Flag("export-jobs").Value.String() == "true"
			exportLogs        = cmd.Flag("export-logs").Value.String() == "true"
			exportProcessLogs = cmd.Flag("export-process-logs").Value.String() == "true"

			exportEnvelopeSizeLimit int64
			exportLogLevel          api.LogLevel

			outputWriter io.WriteCloser = os.Stdout
			showProgress                = false

			wg sync.WaitGroup
		)
		if output != "-" {
			f, err := os.Create(output)
			if err != nil {
				Logger.Err(err).Msg("could not open output file")
			}
			outputWriter = f

			showProgress = !quiet

			if strings.HasSuffix(output, ".yaml") {
				yaml = true
			}
		}
		if yaml {
			r, w := io.Pipe()
			defer w.Close()
			wg.Add(1)
			go func(outputWriter io.WriteCloser) {
				defer wg.Done()
				defer outputWriter.Close()
				br := bufio.NewReader(r)
				first, err := br.Peek(1)
				if err != nil {
					return
				}
				switch first[0] {
				case '[':
					if err := utils.JSON2YAML(outputWriter, br); err != nil {
						Logger.Err(err).Msg("yaml conversion failed")
					}
				case '-':
					if _, err := io.Copy(outputWriter, br); err != nil {
						Logger.Err(err).Msg("error copying the yaml stream")
					}
				default:
					Logger.Error().Msg("Could not identify the content type (not json nor yaml), copying")
					if _, err := io.Copy(outputWriter, br); err != nil {
						Logger.Err(err).Msg("error copying the stream")
					}
				}
			}(outputWriter)
			outputWriter = w
		}

		if err := exportLogLevel.UnmarshalText(
			[]byte(cmd.Flag("export-logs-level").Value.String()),
		); err != nil {
			Logger.Error().Msgf(
				"Invalid log level: '%s'. Got %s",
				cmd.Flag("export-logs-level").Value.String(), err)
			os.Exit(1)
		}

		if flag := cmd.Flag("export-envelope-size-limit"); flag.Changed {
			value, err := strconv.Atoi(flag.Value.String())
			if err != nil {
				Logger.Error().Msgf(
					"Invalid value '%s' for --export-envelope-size-limit: %s",
					flag.Value.String(), err)
				os.Exit(1)
			}
			exportEnvelopeSizeLimit = int64(value)
		}

		c, err := GetClient(false)
		if err != nil {
			return
		}
		filter, err := getProcessFilter(
			c, args, filterFlags{IncludeAllStatuses: true})
		if err != nil {
			Logger.Error().Msgf("Invalid filter: %s", err)
			return
		}
		list, err := getProcessList(c, filter, "", quiet)
		if err != nil {
			Logger.Error().Msgf("Error getting process list: %s", err)
			return
		}
		var (
			pids     api.UUIDSlice
			pidbatch []api.UUIDSlice
		)
		totalCount := int64(len(list))
		for _, p := range list {
			pids = append(pids, p.GetIDAsUUID())
			if len(pids) >= 10000 {
				pidbatch = append(pidbatch, pids)
				pids = nil
			}
		}
		if len(pids) > 0 {
			pidbatch = append(pidbatch, pids)
		}
		ctx, cancel := contextCancelOnSignal(
			context.Background(), os.Interrupt,
		)

		wg.Add(1)
		ch := make(chan []byte, 1000)

		go func() {
			defer wg.Done()
			var (
				started   bool
				lastChunk []byte
			)

			for chunk := range ch {
				if started && chunk == nil {
					// Drop closing ']' of the previous chunk
					lastChunk = bytes.TrimSuffix(lastChunk, []byte("\n]"))
					lastChunk = append(lastChunk, ',', '\n')
				}
				if lastChunk != nil {
					if _, err := outputWriter.Write(lastChunk); err != nil {
						Logger.Err(err).Msg("could not write to file")
						cancel()
						return
					}
				} else if started {
					// first chunk of a new batch
					chunk = bytes.TrimPrefix(chunk, []byte("["))
				}
				lastChunk = chunk
				if chunk != nil && !started {
					started = true
				}
			}

			if lastChunk != nil {
				if _, err := outputWriter.Write(lastChunk); err != nil {
					Logger.Err(err).Msg("could not write to file")
				}
			}

			if err := outputWriter.Close(); err != nil {
				Logger.Err(err).Msg("could not close output")
			}
		}()

		var count int64
		for _, pids := range pidbatch {
			ch <- nil // Signal a new batch
			if err := c.Control.Process.ExportStream(
				ctx,
				&control.ProcessExportRequest{
					ProcessIDs:        pids.ToBytes(),
					ExportEnvelope:    exportEnvelope,
					EnvelopeSizeLimit: exportEnvelopeSizeLimit,
					ExportJobs:        exportJobs,
					ExportLogs:        exportLogs,
					ExportProcessLogs: exportProcessLogs,
					ExportFull:        full,
				},
				func(ctx context.Context, msg *control.OctetStreamList) {
					for _, chunk := range msg.Chunks {
						ch <- chunk.Data
					}
					if showProgress {
						fmt.Printf("\r%d%%", 100*(count+msg.GetProgression())/totalCount)
					}
				},
			); err != nil {
				Logger.Error().Msgf("Error exporting stream: %s", err)
				os.Exit(1)
			}
			count += int64(len(pids))
		}
		if showProgress {
			fmt.Println()
		}
		close(ch)
		wg.Wait()
	},
}

func init() {
	psCmd.AddCommand(psExportCmd)

	flags := psExportCmd.PersistentFlags()
	flags.String("output", "-", "Write the export to this file")
	flags.Bool("yaml", false, "Use the legacy yaml format")
	flags.Bool("export-envelope", false, "Export the process envelopes")
	flags.Int64("export-envelope-size-limit", 0,
		"When exporting the envelopes, each envelope will be exported up to the given limit in bytes")
	flags.Bool("export-jobs", false, "Export the process jobs")
	flags.Bool("export-logs", false, "Export the actor logs associated to the process")
	flags.String("export-logs-level", "notice", "Log level when exporting actor logs ('notice', 'warning', 'error')")
	flags.Bool("export-process-logs", false, "Export the process logs")

	flags.Bool("full", false, "Export all the optional data (See also the --export-xxx options)")
	flags.Bool("quiet", false, "Disable extra logging so stdout only contains the export")
}
