// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"context"
	"os"

	"github.com/spf13/cobra"
	"xbus.io/go-xbus/v4/api"

	"xbus.io/xbus/storage"
)

const (
	resetCertArgKey = iota
	resetCertArgCount
)

var accountResetCertCmd = &cobra.Command{
	Use:   "reset-cert <id|name>",
	Short: "Un-accept an account so it can reset its cert",
	Run:   resetCert,
}

func resetCert(cmd *cobra.Command, args []string) {
	if len(args) < resetCertArgCount {
		Logger.Error().Msgf("Not enough arguments; %d arguments are required.", resetCertArgCount)
		if err := cmd.Usage(); err != nil {
			Logger.Error().Msgf("Failed to print usage: %s", err)
		}
		os.Exit(1)
	}

	os.Exit(doResetCert(MakeCmdContext(), getAccountStorage(), args[resetCertArgKey]))
}

func doResetCert(ctx context.Context, accountStorage storage.AccountStorage, identifier string) int {
	account, found := findAccount(ctx, accountStorage, identifier)
	if !found {
		return 1
	}

	if account.Status == api.Account_PENDING {
		Logger.Error().Msgf(
			"The account is already in pending state. " +
				"It must be accepted or rejected first")
		return 1
	}

	if account.Status != api.Account_ENABLED && account.Status != api.Account_DISABLED {
		Logger.Error().Msgf(
			"The account current state does now allow a cert reset. "+
				"Expected ENABLED or DISABLED, got %s",
			account.Status)
		return 1
	}

	account.Status = api.Account_PENDING
	account.CSR = ""

	var err error
	_, err = accountStorage.Update(ctx, account)

	if err != nil {
		Logger.Error().Msgf("Error saving the account: %s", err)
		return 1
	}
	return 0
}

func init() {
	accountCmd.AddCommand(accountResetCertCmd)
}
