// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package cmd

import (
	"os"
	"time"

	"github.com/spf13/cobra"
	"github.com/spf13/viper"

	"xbus.io/xbus/utils"
)

// certGenerateClientCACmd represents the client CA generation sub-command.
var certGenerateClientCACmd = &cobra.Command{
	Use:   "client-ca",
	Short: "Generate certificates for the client CA.",
	Long: "Generate certificates for the client CA (the one clients connecting to Xbus will " +
		"have to have been signed by). The file will be named \"ClientCA.crt\" and is based on " +
		"a \"ClientCA.key\" private key.",
	Run: generateClientCA,
}

// generateClientCA parses arguments to generate the client CA.
func generateClientCA(cmd *cobra.Command, _ []string) {
	certDir := viper.GetString("cert-dir")
	expire, err := getExpire(cmd, "expire")
	if err != nil {
		Logger.Error().Msgf("Invalid --expire: %s", err)
		return
	}
	hostDefs := cmd.Flag(flagHostDefs).Value.String()
	if hostDefs == "" {
		hostDefs = viper.GetString("nats.host")
	}
	os.Exit(doGenerateClientCA(certDir, expire, hostDefs))
}

// doGenerateClientCA generates the client CA.
func doGenerateClientCA(certDir string, expire time.Time, hostDefs string) int {
	generated := genCert(
		certDir+"/ClientCA.crt", certDir+"/ClientCA.key",
		func(priv interface{}) ([]byte, error) {
			return utils.CreateClientCA(priv, time.Now(), expire, hostDefs)
		},
	)
	if !generated {
		return 1
	}
	return 0
}

func init() {
	certGenerateCmd.AddCommand(certGenerateClientCACmd)

	flags := certGenerateClientCACmd.Flags()
	flags.String(flagHostDefs, "", flagHostDefsHelp)
	flags.String(flagExpire, "1y", flagExpireHelp)
}
