// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package fullstack

import (
	"context"
	"fmt"
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
	xbus "xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/api"

	"xbus.io/xbus/services/envelopestorage"
	"xbus.io/xbus/testutils"
)

func TestProcessCancelOnStalledEnvelope(t *testing.T) {
	defer switchDurationVar(&envelopestorage.EnvelopeStallDelay, 1*time.Second)()
	ft := InitFullstack(t, true)
	defer ft.Shutdown()

	ft.CreateRole(api.Actor_EMITTER, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	assert.Equal(t, "bob", emitter.Name)

	canceled := make(chan bool, 1)

	ft.CreateRole(api.Actor_CONSUMER, "anyconsumer")

	ctx, cancel := context.WithTimeout(context.Background(), 4*time.Second)

	defer cancel()

	_, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})
	xsub, err := consumer.Subscribe(
		func(apc *xbus.ActorProcessingContext) error {
			t.Logf("Start reading the envelope so the stall is detected")

			envelope, err := apc.GetEnvelope("default")
			if err != nil {
				return err
			}

			select {
			case <-apc.Context.Done():
				canceled <- true
			case <-time.After(2 * time.Second):
				canceled <- false
			}
			_ = envelope.CloseWithError(fmt.Errorf("stop reading"))

			ft.Logger.Info().Msgf("consumer Process end")

			return nil
		})
	assert.Nil(t, err)
	defer testutils.DeferedAssertNoError(t, xsub.Unsubscribe)

	ft.InstallGraph(BuildSimpleGraph("evt1", "anyconsumer"))
	env := xbus.ForgeEnvelopeFromItems("evt1", []api.Item{[]byte("\"item1\"")})
	env.Last = false
	env.Events[0].ItemCount = 2

	envStateChan, sub := ft.GetEnvStateChan(emitter, api.UUID{})
	defer testutils.DeferedAssertNoError(t, sub.Unsubscribe)

	status, err := emitter.SendEnvelope(nil, api.DefaultOutput, env)
	assert.Nil(t, err)
	assert.Equal(t, api.EnvelopeAck_RECEIVING, status)

	// The second item will never be sent. In about 15 seconds, the envelope
	// should be marked as STALLED, and the process canceled.

	select {
	case <-ctx.Done():
		t.Errorf("Process state not updated (expecting change to RUNNING)")
		return
	case envState := <-envStateChan:
		ft.Logger.Info().Msgf("Process is RUNNING as expected")
		assert.Equal(t, api.Process_RUNNING, envState.Status)
	}

	select {
	case success := <-canceled:
		if !success {
			t.Errorf("The processing was not canceled")
		}
		ft.Logger.Info().Msgf("The actor process was canceled as expected")

	case <-ctx.Done():
		t.Errorf("Cancel failed")
	}

	select {
	case <-ctx.Done():
		t.Errorf("Process state not updated (expecting change to ERROR)")
		return
	case envState := <-envStateChan:
		assert.Equal(t, api.Process_ERROR, envState.Status)
		ft.Logger.Info().Msgf("Process is ERROR as expected")
	}
}
