// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package broker

import (
	"context"
	"time"

	nrpc "github.com/nats-rpc/nrpc"
	"xbus.io/go-xbus/v4/api"

	"xbus.io/xbus/process"
	"xbus.io/xbus/services"
	"xbus.io/xbus/storage"
)

const (
	// BrokerQueue is the name we use to queueSub to a queue with the broker
	BrokerQueue = "broker_queue"

	// DefaultPoolSize is the default size of the worker pool
	DefaultPoolSize = 5

	// DefaultPoolMaxPending is the default number of requests that can be
	// pending
	DefaultPoolMaxPending = 5

	// DefaultPoolMaxPendingDuration is the default max pending duration for a
	// a request
	DefaultPoolMaxPendingDuration = time.Second
)

// Options is used to create a Service
type Options struct {
	services.Options

	ActorStorage    storage.ActorStorage
	EnvelopeStorage storage.EnvelopeStorage
	LogStorage      storage.LogStorage
	ProcessManager  *process.Manager

	PoolSize               uint
	PoolMaxPending         uint
	PoolMaxPendingDuration time.Duration
}

// NewService returns a new Service
func (o Options) NewService(ctx context.Context) *Service {
	if o.PoolSize == 0 {
		o.PoolSize = DefaultPoolSize
	}
	if o.PoolMaxPending == 0 {
		o.PoolMaxPending = DefaultPoolMaxPending
	}
	if o.PoolMaxPendingDuration == 0 {
		o.PoolMaxPendingDuration = DefaultPoolMaxPendingDuration
	}
	return &Service{
		BaseService:    services.NewBaseService(ctx, "broker", o.Options),
		Opts:           o,
		processManager: o.ProcessManager,
	}
}

// New returns a new Service
func New(ctx context.Context, options Options) *Service {
	return options.NewService(ctx)
}

// Service is the central node of all envelope exchanges
type Service struct {
	services.BaseService
	Opts Options

	processManager *process.Manager

	workers       *nrpc.WorkerPool
	brokerHandler *api.BrokerHandler
}

// Startup initialize all the subscriptions
func (s *Service) Startup() error {
	if err := s.BaseService.Startup(); err != nil {
		return err
	}
	s.workers = nrpc.NewWorkerPool(
		s.Ctx, s.Opts.PoolSize, s.Opts.PoolMaxPending, s.Opts.PoolMaxPendingDuration,
	)

	s.brokerHandler = api.NewBrokerConcurrentHandler(
		s.workers, s.Conn, brokerServer{s},
	)

	return s.SubscribeAll(
		s.HandlerQueueSubscribe(s.brokerHandler, BrokerQueue),
	)
}
