// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package clientapi

import (
	"context"
	"fmt"
	"time"

	nrpc "github.com/nats-rpc/nrpc"
	"google.golang.org/protobuf/types/known/timestamppb"
	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/tools"

	"xbus.io/xbus/services"
	"xbus.io/xbus/storage"
)

// QueueName is the queue name used by the service
const QueueName = "xbus.services.clientapi"

// Options for the Service
type Options struct {
	services.Options

	ActorStorage   storage.ActorStorage
	AccountStorage storage.AccountStorage
	SessionStorage storage.SessionStorage
}

// NewService instantiate a Service
func NewService(ctx context.Context, o Options) *Service {
	return &Service{
		o:           o,
		BaseService: services.NewBaseService(ctx, "clientapi", o.Options),
	}
}

// Service provides an api for administrators
type Service struct {
	services.BaseService

	o Options

	clientAPIHandler *api.ClientAPIHandler
}

// Startup starts the service
func (s *Service) Startup() error {
	if err := s.BaseService.Startup(); err != nil {
		return err
	}
	whoamiSubject, _ := api.WhoAmISubjects("*")
	s.clientAPIHandler = api.NewClientAPIHandler(
		s.Ctx, s.Conn, clientAPIServer{s},
	)
	whoamiHandler := api.NewWhoAmIHandler(
		s.Ctx, s.Conn, clientAPIServer{s},
	)
	// The two handlers have the same subject prefix 'xbus.client',
	// so we need to subscribe to their methods directly
	// TODO the api need some rework to avoid this situation
	if err := s.SubscribeAll(
		services.QueueSubscribe(
			api.XbusAPIPrefix+"client.*.getactors", QueueName,
			s.clientAPIHandler.Handler,
		),
		services.QueueSubscribe(
			api.XbusAPIPrefix+"client.*.getactors.*", QueueName,
			s.clientAPIHandler.Handler,
		),
		services.QueueSubscribe(
			api.XbusAPIPrefix+"client.*.getsessiontoken", QueueName,
			s.clientAPIHandler.Handler,
		),
		services.QueueSubscribe(
			api.XbusAPIPrefix+"client.*.getsessiontoken.*", QueueName,
			s.clientAPIHandler.Handler,
		),
		services.QueueSubscribe(
			whoamiSubject, QueueName,
			whoamiHandler.Handler),
		services.QueueSubscribe(
			whoamiSubject+".json", QueueName,
			whoamiHandler.Handler),
	); err != nil {
		_ = s.Shutdown()
		return err
	}
	return nil
}

type clientAPIServer struct {
	*Service
}

func (s clientAPIServer) WhoAmI(
	ctx context.Context, identifier string,
) (*api.Account, error) {
	var (
		account *api.Account
		ok      bool
	)

	if id, isUUID := api.UUIDFromString(identifier); isUUID == nil {
		account, ok = s.o.AccountStorage.Get(ctx, id)
		if !ok {
			panic("No account by this id")
		}
	} else {
		account, ok = s.o.AccountStorage.GetByName(ctx, identifier)
		if !ok {
			account, ok = s.o.AccountStorage.GetByAPIKey(ctx, identifier)
			if !ok {
				panic("No account by this name or api key")
			}
		}
	}
	return account, nil
}

func (s clientAPIServer) GetActors(ctx context.Context) (
	*api.ClientGetActorsReply, error,
) {
	accountSID := nrpc.GetRequest(ctx).ServiceParam("accountID")
	accountID, err := api.UUIDFromString(accountSID)
	if err != nil {
		return nil, err
	}

	return &api.ClientGetActorsReply{
		Actors: s.o.ActorStorage.ListAccount(ctx, accountID),
	}, nil
}

func (s clientAPIServer) GetSessionToken(
	ctx context.Context, request *api.GetSessionTokenRequest,
) (*api.SessionToken, error) {
	accountSID := nrpc.GetRequest(ctx).ServiceParam("accountID")
	accountID, err := api.UUIDFromString(accountSID)
	if err != nil {
		return nil, err
	}
	if request.InvalidateToken != "" {
		session, err := s.o.SessionStorage.Get(ctx, request.InvalidateToken)
		if err != nil {
			s.Log.Err(err).Msg("error invalidating session token")
		}
		if session != nil {
			if accountID != session.AccountID {
				return nil, fmt.Errorf("InvalidateToken value is invalid (trying to invalidate another account session?)")
			}
		}
	}

	// Max session duration is 12h (TODO make it setting)
	expiresIn := request.ExpiresIn
	if expiresIn > 60*60*12 {
		expiresIn = 60 * 60 * 12
	}

	newToken := api.SessionToken{
		Token:      tools.RandomString(32),
		ValidUntil: timestamppb.New(time.Now().Add(time.Duration(expiresIn) * time.Second).UTC()),
	}

	if err := s.o.SessionStorage.Set(ctx, storage.Session{
		SessionToken: &newToken,
		AccountID:    accountID,
	}); err != nil {
		s.Log.Err(err).Msg("error saving new session")
		return nil, fmt.Errorf("Internal Error")
	}
	// Now that we secured the creation of a new token, we can delete the old one

	if err := s.o.SessionStorage.Delete(ctx, request.InvalidateToken); err != nil {
		s.Log.Err(err).
			Str("token", request.InvalidateToken).
			Msg("could not delete the session")
	}
	return &newToken, nil
}
