// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package control

import (
	"context"
	"crypto/x509"
	"encoding/pem"
	"errors"
	"fmt"
	"math/rand"
	"time"

	xbus "xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/api/control"
	"xbus.io/go-xbus/v4/tools"

	"xbus.io/xbus/storage"
)

type accountServer struct {
	*Service
}

var letters = []rune("0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ")

func randSeq(n int) string {
	b := make([]rune, n)
	for i := range b {
		b[i] = letters[rand.Intn(len(letters))]
	}
	return string(b)
}

// AccountAccept accepts a pending account request; a certificate is generated from the CSR.
func AccountAccept(
	ctx context.Context,
	accountStorage storage.AccountStorage, accountID api.UUID,
	serverCert *x509.Certificate, serverKey interface{},
	expire time.Time,
) (*api.Account, error) {
	a, exists := accountStorage.Get(ctx, accountID)
	if !exists {
		return nil, fmt.Errorf("No such account: %s", accountID)
	}
	if a.Status != api.Account_PENDING {
		return nil, fmt.Errorf("Invalid initial status %s for 'Accept'", a.Status)
	}
	p, r := pem.Decode([]byte(a.CSR))
	if len(r) != 0 {
		return nil, fmt.Errorf("Invalid CSR pem encoding")
	}
	if p.Type != "CERTIFICATE REQUEST" {
		return nil, fmt.Errorf("Invalid CSR pem encoding")
	}
	cert, err := tools.CreateCertificateFromCSR(
		p.Bytes, serverCert, serverKey, expire,
	)
	if err != nil {
		return nil, err
	}
	a.Cert = string(pem.EncodeToMemory(&pem.Block{
		Type:  "CERTIFICATE",
		Bytes: cert,
	}))
	a.CSR = ""
	a.Status = api.Account_ENABLED
	return accountStorage.Update(ctx, a)
}

// AccountReject rejects a pending account request.
func AccountReject(
	ctx context.Context, accountStorage storage.AccountStorage, accountID api.UUID,
) (*api.Account, error) {
	a, exists := accountStorage.Get(ctx, accountID)
	if !exists {
		return nil, fmt.Errorf("No such account: %s", accountID)
	}
	if a.Status != api.Account_PENDING {
		return nil, fmt.Errorf("Invalid initial status %q for 'Reject'", a.Status)
	}
	a.CSR = ""
	a.Status = api.Account_DISABLED
	return accountStorage.Update(ctx, a)
}

func accountDelete(
	ctx context.Context,
	storage storage.AccountStorage,
	actorStorage storage.ActorStorage,
	accountID api.UUID,
) error {
	a, exists := storage.Get(ctx, accountID)
	if !exists {
		return xbus.ErrNotFound
	}
	for _, actor := range actorStorage.ListAccount(ctx, a.GetIDAsUUID()) {
		actorStorage.Delete(ctx, actor.GetIDAsUUID())
	}
	storage.Delete(ctx, a)
	return nil
}

func (s accountServer) Delete(
	ctx context.Context, req *control.AccountIDRequest,
) (err error) {
	err = accountDelete(ctx, s.o.AccountStorage, s.o.ActorStorage, req.GetAccountIDAsUUID())
	if err != nil {
		if !errors.Is(err, xbus.ErrNotFound) {
			err = api.ClientProtoError("No such account")
		} else {
			err = api.ServerProtoError(err)
		}
		return
	}
	return
}

func (s accountServer) List(
	ctx context.Context, _ *control.AccountListRequest,
) (*control.AccountListReply, error) {
	resp := control.AccountListReply{
		AccountList: s.o.AccountStorage.List(ctx),
	}
	return &resp, nil
}

// HasPendingCSR returns true if the given account has a CSR pending
func HasPendingCSR(account *api.Account) bool {
	return account.Status == api.Account_PENDING && account.CSR != ""
}

func (s accountServer) Update(
	ctx context.Context, req *control.AccountUpdateRequest,
) (
	resp *api.Account, err error,
) {
	if req.Account.GetIDAsUUID().IsZero() {
		err = fmt.Errorf("Must pass an ID")
		return
	}
	a, exists := s.o.AccountStorage.Get(ctx, req.Account.GetIDAsUUID())
	if !exists {
		err = fmt.Errorf("No such account: %s", req.Account.ID)
		return
	}
	switch {
	case HasPendingCSR(a) && req.Account.Status == api.Account_ENABLED:
		a, err = AccountAccept(
			ctx, s.o.AccountStorage, req.Account.GetIDAsUUID(),
			s.o.ServerCert, s.o.ServerKey, req.Expire.AsTime())
	case HasPendingCSR(a) && req.Account.Status == api.Account_DISABLED:
		a, err = AccountReject(ctx, s.o.AccountStorage, req.Account.GetIDAsUUID())
	default:
		err = fmt.Errorf("Invalid change")
		return
	}
	if err != nil {
		return
	}
	resp = &api.Account{
		ID:     a.ID,
		Name:   a.Name,
		Type:   a.Type,
		Status: a.Status,
		CSR:    a.CSR,
		Cert:   a.Cert,
	}
	return
}

func (s accountServer) RenewAPIKey(
	ctx context.Context, req *control.AccountIDRequest,
) (
	account *api.Account, err error,
) {
	account, ok := s.o.AccountStorage.Get(ctx, req.GetAccountIDAsUUID())
	if !ok {
		err = fmt.Errorf("No such account")
		return
	}

	account.ApiKey = randSeq(32)
	account, err = s.o.AccountStorage.Update(ctx, account)
	if err != nil {
		err = api.ServerProtoError(err)
	}
	return
}

func (s accountServer) GetAPIKey(
	ctx context.Context, req *control.AccountIDRequest,
) (
	account *api.Account, err error,
) {
	account, ok := s.o.AccountStorage.Get(ctx, req.GetAccountIDAsUUID())
	if !ok {
		err = fmt.Errorf("No such account")
		return
	}
	return
}

func (s accountServer) SetAPIKey(
	ctx context.Context, req *control.AccountUpdateRequest,
) (
	account *api.Account, err error,
) {
	account, ok := s.o.AccountStorage.Get(ctx, req.Account.GetIDAsUUID())
	if !ok {
		err = fmt.Errorf("No such account")
		return
	}

	account.ApiKey = req.Account.ApiKey
	account, err = s.o.AccountStorage.Update(ctx, account)
	if err != nil {
		err = api.ServerProtoError(err)
	}
	return
}

func (s accountServer) ResetCert(
	ctx context.Context, req *control.AccountIDRequest,
) (
	err error,
) {
	account, ok := s.o.AccountStorage.Get(ctx, req.GetAccountIDAsUUID())
	if !ok {
		err = fmt.Errorf("No such account")
		return
	}

	if account.Status == api.Account_PENDING {
		return fmt.Errorf("The account is already in pending state. It must be accepted or rejected first")
	}

	if account.Status != api.Account_ENABLED && account.Status != api.Account_DISABLED {
		return fmt.Errorf("The account current state does now allow a cert reset. Expected ENABLED or DISABLED, got %s", account.Status)
	}

	account.Status = api.Account_PENDING
	account.CSR = ""

	_, err = s.o.AccountStorage.Update(ctx, account)

	return err
}

func (s accountServer) PasswordSet(
	ctx context.Context, req *control.AccountPasswordSetRequest,
) (
	err error,
) {
	_, ok := s.o.AccountStorage.Get(ctx, req.GetAccountIDAsUUID())
	if !ok {
		err = fmt.Errorf("No such account")
		return
	}

	if req.Password == "" {
		s.o.PasswordStorage.Reset(ctx, req.GetAccountIDAsUUID())
	} else {
		s.o.PasswordStorage.Set(ctx, req.GetAccountIDAsUUID(), req.Password)
	}
	return nil
}
