// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

// Package core provides the 'core' service, which must be started once and only once
// in every server process
package core

import (
	"context"

	"google.golang.org/protobuf/types/known/timestamppb"
	"xbus.io/go-xbus/v4/api"

	"xbus.io/xbus/pipeline"
	"xbus.io/xbus/process"
	"xbus.io/xbus/services"
	"xbus.io/xbus/storage"
)

// Options is used to create a 'core' service
type Options struct {
	services.Options

	ActorStorage    storage.ActorStorage
	ProcessStorage  storage.ProcessStorage
	ProcessManager  *process.Manager
	PipelineManager *pipeline.Manager
	LogStorage      storage.LogStorage
	EnvelopeStorage storage.EnvelopeStorage
}

// NewService creates a Emitter
func (o Options) NewService(ctx context.Context) *Service {
	return &Service{
		BaseService: services.NewBaseService(ctx, "core", o.Options),
		o:           o,
	}
}

// Service is in charge of emitting envelopes to Actors
type Service struct {
	services.BaseService

	processLogEntries chan *processLogEntries

	o Options

	coreEvents *api.CoreEventsHandler
}

// Startup starts the service
func (s *Service) Startup() error {
	if err := s.BaseService.Startup(); err != nil {
		return err
	}
	s.coreEvents = api.NewCoreEventsHandler(s.Ctx, s.Conn, nil)
	s.processLogEntries = make(chan *processLogEntries, 10)
	if err := s.initProcessLogEntryHandler(); err != nil {
		close(s.processLogEntries)
		return err
	}
	s.o.ActorStorage.SetEventHandlers(s.onActorUp, s.onActorDown)
	s.o.ProcessStorage.SetEventHandlers(s.onProcessStatusChange)
	s.o.ProcessStorage.SetProcessLogEntryHandler(s.processLogEntryHandler)
	s.o.EnvelopeStorage.SetEventHandler(s.onEnvelope)
	return nil
}

// Shutdown stops the service
func (s *Service) Shutdown() error {
	s.o.ActorStorage.SetEventHandlers(nil, nil)
	s.o.ProcessStorage.SetEventHandlers(nil)
	s.o.ProcessStorage.SetProcessLogEntryHandler(nil)
	s.o.EnvelopeStorage.SetEventHandler(nil)
	close(s.processLogEntries)
	return s.BaseService.Shutdown()
}

func (s *Service) onActorUp(actor *api.Actor) {
	if err := s.coreEvents.ActorPublish(&api.CoreEvent{
		Name: "up",
		Actor: &api.Actor{
			ID:    actor.ID,
			Name:  actor.Name,
			Kind:  actor.Kind,
			Roles: actor.Roles,
		},
	}); err != nil {
		s.o.Logger.Err(err).
			Str("actor", actor.Name).
			Msg("Failed to publish actor event 'up'")
	}
}

func (s *Service) onActorDown(actor *api.Actor) {
	if err := s.coreEvents.ActorPublish(&api.CoreEvent{
		Name: "down",
		Actor: &api.Actor{
			ID:    actor.ID,
			Name:  actor.Name,
			Kind:  actor.Kind,
			Roles: actor.Roles,
		},
	}); err != nil {
		s.o.Logger.Err(err).
			Str("actor", actor.Name).
			Msg("Failed to publish actor event 'down'")
	}
}

func (s *Service) onProcessStatusChange(process storage.Process, oldStatus api.Process_Status, status api.Process_Status) {
	if err := s.coreEvents.ProcessPublish(
		&api.CoreEvent{
			Name: "status",
			Process: &api.Process{
				ID:                process.ID.ToBytes(),
				CreatedAt:         timestamppb.New(process.CreatedAt),
				TriggerEmitterID:  process.TriggerEmitterID.ToBytes(),
				TriggerEnvelopeID: process.TriggerEnvelopeID.ToBytes(),
				PipelineID:        process.PipelineID.ToBytes(),
				Status:            status,
			},
			ProcessOldStatus: oldStatus,
		},
	); err != nil {
		s.Log.Err(err).Msg("Error publishing ProcessEvent")
	}
}

func (s *Service) onNodesReady(event *api.ProcessNodeReadyEvent) {
	if err := s.coreEvents.ProcessNodeReadyPublish(event); err != nil {
		s.Log.Err(err).Msg("Error publising ProcessNodeReady")
	}
}

func (s *Service) onEnvelope(id api.UUID, status storage.EnvelopeStorageReceptionStatus, newData bool) {
	// TODO make EnvelopePublish parametrizable with subject args so a
	// subscriber can listen to specific envelopes. This means supporting
	// subject params on request-less nrpc functions
	if err := s.coreEvents.EnvelopePublish(&api.EnvelopeEvent{
		EnvelopeID: id.ToBytes(),
		Status:     api.EnvelopeEvent_Status(status), // We know the enums are compatible
		NewData:    newData,
	}); err != nil {
		s.Log.Err(err).Msg("Error publising Envelope event")
	}
}
