// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package registration

import (
	"context"
	"encoding/pem"
	"fmt"

	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/tools"

	"xbus.io/xbus/services"
	"xbus.io/xbus/storage"
)

// QueueName used for queue subscriptions
const QueueName = "xbus.services.registration"

// Options of the Service
type Options struct {
	services.Options

	AccountStorage storage.AccountStorage
	ActorStorage   storage.ActorStorage
	ServerCA       []byte
}

// NewService instantiate a Service
func NewService(ctx context.Context, o Options) *Service {
	return &Service{
		o:           o,
		BaseService: services.NewBaseService(ctx, "registration", o.Options),
	}
}

// Service implements the registration api
type Service struct {
	services.BaseService
	o Options
}

// Startup subscribes the service to the registration apis
func (s *Service) Startup() error {
	if err := s.BaseService.Startup(); err != nil {
		return err
	}
	regHandler := api.NewSvcRegistrationHandler(
		s.Ctx, s.Conn, registrationServer{s},
	)
	if err := s.SubscribeAll(
		s.HandlerQueueSubscribe(regHandler, QueueName),
	); err != nil {
		_ = s.Shutdown()
		return err
	}
	return nil
}

type registrationServer struct {
	*Service
}

func (s registrationServer) Register(
	ctx context.Context, req *api.RegistrationRequest,
) (*api.Registration, error) {
	return s.handleRegistration(ctx, api.UUID{}, req)
}

func (s registrationServer) Update(
	ctx context.Context, accountID string, req *api.RegistrationRequest,
) (*api.Registration, error) {
	accountUUID, err := api.UUIDFromString(accountID)
	if err != nil {
		return nil, fmt.Errorf("invalid accountID: %w", err)
	}
	return s.handleRegistration(ctx, accountUUID, req)
}

func (s *Service) checkActors(
	ctx context.Context, accountID api.UUID, actors []*api.Actor,
) error {
	// For each actor, if not-yet created (ie if this is an original req),
	// create the actor in the pending state
	for _, actor := range actors {
		if a, exists := s.o.ActorStorage.GetByName(ctx, actor.Name); exists {
			if accountID.IsZero() || a.GetAccountIDAsUUID() != accountID {
				return fmt.Errorf(
					"%s: Actor %s is attached to a different account (%s)",
					accountID, a.Name, a.GetAccountIDAsUUID().String())
			}
		}
	}
	return nil
}

func (s *Service) handleRegistration(
	ctx context.Context, accountID api.UUID, req *api.RegistrationRequest,
) (*api.Registration, error) {
	var account *api.Account

	if accountID.IsZero() {
		var (
			ok  bool
			err error
		)
		if !tools.ValidateAccountName(req.Name) {
			return nil, fmt.Errorf("Invalid account name: '%s'", req.Name)
		}
		account, ok = s.o.AccountStorage.GetByName(ctx, req.Name)
		if !ok {
			if err := s.checkActors(ctx, accountID, req.ActorList); err != nil {
				return nil, err
			}
			// If non-existing account, create it in the adequate state and its csr attached
			account = &api.Account{
				Name:   req.Name,
				Type:   req.Type,
				Status: api.Account_PENDING,
			}
			if req.CSR != "" {
				account.CSR = req.CSR
				account.CSROrigin = api.Account_ANONYMOUS
			}
			account, err = s.o.AccountStorage.Create(ctx, account)
			if err != nil {
				return nil, api.ServerProtoError(err)
			}
		} else {
			if err := s.checkActors(ctx, account.GetIDAsUUID(), req.ActorList); err != nil {
				return nil, err
			}
			if account.CSR != "" && req.CSR != account.CSR {
				return nil, fmt.Errorf("CSR validation already pending. Accept or reject first")
			}
			if account.CSR == "" && account.Status == api.Account_PENDING {
				account.CSR = req.CSR
				account.CSROrigin = api.Account_ANONYMOUS
				var err error
				account, err = s.o.AccountStorage.Update(ctx, account)
				if err != nil {
					return nil, err
				}
			}
		}
	} else {
		var exists bool
		account, exists = s.o.AccountStorage.Get(ctx, accountID)
		if !exists {
			return nil, api.ClientProtoError(
				fmt.Sprintf("No such account %s", accountID),
			)
		}
		if err := s.checkActors(ctx, account.GetIDAsUUID(), req.ActorList); err != nil {
			return nil, err
		}
		if req.CSR != "" {
			if account.Status != api.Account_PENDING {
				return nil, fmt.Errorf("To renew an account cert, you must first run 'xbusctl account reset-cert <account>'")
			}
			if account.CSR != "" && account.CSR != req.CSR {
				return nil, fmt.Errorf("CSR validation already pending. Accept or reject first")
			}
			account.CSR = req.CSR
			account.CSROrigin = api.Account_AUTHACCOUNT
			var err error
			account, err = s.o.AccountStorage.Update(ctx, account)
			if err != nil {
				return nil, fmt.Errorf("failed to update account: %w", err)
			}
		}
	}

	// For each actor, if not-yet created (ie if this is an original req),
	// create the actor in the pending state
	for _, actor := range req.ActorList {
		if _, exists := s.o.ActorStorage.GetByName(ctx, actor.Name); !exists {
			sActor := api.Actor{
				Name:      actor.Name,
				Kind:      actor.Kind,
				AccountID: account.ID,
				Status:    api.Actor_PENDING,
				Roles:     actor.Roles,
			}
			var err error
			_, err = s.o.ActorStorage.Create(ctx, &sActor)
			if err != nil {
				return nil, api.ClientProtoError(
					"Error creating Actor '%s': %s",
					actor.Name, err,
				)
			}
		}
	}

	// Check the current state of the account and actors and put it all
	// in the reply, along with the signed certificate if applicable
	reg := api.Registration{
		ID: account.ID,
	}
	switch account.Status {
	case api.Account_PENDING:
		reg.RegistrationStatus = api.Registration_PENDING
	case api.Account_ENABLED:
		reg.RegistrationStatus = api.Registration_ACCEPTED
	case api.Account_DISABLED:
		if account.Cert == "" {
			reg.RegistrationStatus = api.Registration_REJECTED
		} else {
			reg.RegistrationStatus = api.Registration_ACCEPTED
		}
	}
	if reg.RegistrationStatus == api.Registration_ACCEPTED {
		reg.SignedCertificate = account.Cert
		if len(s.o.ServerCA) != 0 {
			reg.ServerCA = string(pem.EncodeToMemory(&pem.Block{
				Type:  "CERTIFICATE",
				Bytes: s.o.ServerCA,
			}))
		}
	}

	actorList := s.o.ActorStorage.ListAccount(ctx, account.GetIDAsUUID())

	for _, actor := range actorList {
		reg.ActorList = append(reg.ActorList, &api.Actor{
			ID:     actor.ID,
			Name:   actor.Name,
			Kind:   actor.Kind,
			Status: actor.Status,
			Roles:  actor.Roles,
		})
	}
	return &reg, nil
}
