// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package sentinel

import (
	"context"
	"time"

	"xbus.io/go-xbus/v4/api"

	"xbus.io/xbus/services"
	"xbus.io/xbus/storage"
)

const (
	// ServiceName name of the service
	ServiceName = "sentinel"

	// QueueName queue name for queuesubscribes
	QueueName = "xbus.services." + ServiceName
)

// Settings holds the configuration settings set by the configuration file
// or command line
type Settings struct {
	SurveyOnlineStatusDelay time.Duration
}

// Options is used to create a 'processstate' service
type Options struct {
	services.Options
	Settings

	ActorStorage storage.ActorStorage
}

// NewService creates a Service
func (o Options) NewService(ctx context.Context) *Service {
	return &Service{
		BaseService: services.NewBaseService(ctx, ServiceName, o.Options),
		o:           o,
	}
}

// Service is in charge of emitting process status events to various subjects
type Service struct {
	services.BaseService

	o       Options
	handler *api.SentinelHandler
}

// Startup starts the service
func (s *Service) Startup() error {
	if err := s.BaseService.Startup(); err != nil {
		return err
	}
	sentinelHandler := api.NewSentinelHandler(s.Ctx, s.Conn, sentinelServer{s})
	if err := s.SubscribeAll(
		s.HandlerQueueSubscribe(sentinelHandler, QueueName),
	); err != nil {
		return err
	}
	s.handler = sentinelHandler

	// Ping all actors so they report their status if online
	// XXX Actors that do not respond within 5 seconds may be forced offline
	// but it is less backward compatible. We may do it later.
	if err := s.handler.SurveyPublish(&api.SurveyRequest{OnlineStatus: true}); err != nil {
		s.o.Options.Logger.Err(err).Msg("failed to send the actor survey")
	}

	if s.o.SurveyOnlineStatusDelay != 0 {
		s.TaskAdd()
		go s.run()
	}
	return nil
}

func (s *Service) Shutdown() error {
	return s.BaseService.Shutdown()
}

func (s *Service) run() {
	defer s.TaskDone()

	for {
		select {
		case <-time.After(s.o.SurveyOnlineStatusDelay):
			s.o.Options.Logger.Debug().Msg("running periodic onlinestatus survey")
			if err := s.handler.SurveyPublish(&api.SurveyRequest{OnlineStatus: true}); err != nil {
				s.o.Options.Logger.Err(err).Msg("failed to send the actor survey")
			}

		case <-s.Ctx.Done():
			return
		}
	}
}

type sentinelServer struct {
	*Service
}

func (s sentinelServer) ActorReady(ctx context.Context, actorID string) {
	actorUUID, err := api.UUIDFromString(actorID)
	if err != nil {
		s.Log.Err(err).
			Str("actor-id", actorID).
			Msg("received invalid actorID")
		return
	}
	// signal the actor is online
	if err := s.o.ActorStorage.SetOnline(ctx, actorUUID, true); err != nil {
		s.Log.Err(err).Msg("error calling SetOnline")
	}
}

func (s sentinelServer) ActorLeaving(ctx context.Context, actorID string) {
	actorUUID, err := api.UUIDFromString(actorID)
	if err != nil {
		s.Log.Err(err).
			Str("actor-id", actorID).
			Msg("received invalid actorID")
		return
	}
	// signal the actor is online
	if err := s.o.ActorStorage.SetOnline(ctx, actorUUID, false); err != nil {
		s.Log.Err(err).Msg("error calling SetOnline")
	}
}

// ListActors returns the list of actors.
//
// When returning, we know that the
// service finished any ongoing discovery of the actors statuses.
// The initial implementation is way too naïve, but it
// is not a big deal until we go really distributed (mainly because of the
// director service that rely on it to properly start).
func (s sentinelServer) ListActors(
	ctx context.Context, send func(*api.ClientGetActorsReply),
) error {
	actors := s.o.ActorStorage.List(ctx)

	batchSize := 100

	for batchSize < len(actors) {
		send(&api.ClientGetActorsReply{
			Actors: actors[0:batchSize:batchSize],
		})
		actors = actors[batchSize:]
	}
	send(&api.ClientGetActorsReply{
		Actors: actors,
	})
	return nil
}
