// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package tools

import (
	"errors"
	"sync"

	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/api/control"
)

// ErrNoSuchActor is returned if the pipeline does not exist
var ErrNoSuchActor = errors.New("no such pipeline")

// NewActorCache creates a ActorCache instance
func NewActorCache(control *control.Client) *ActorCache {
	return &ActorCache{
		control: control,
		cache:   make(map[api.UUID]*api.Actor),
	}
}

// ActorCache caches the pipeline definitions
type ActorCache struct {
	control *control.Client
	cache   map[api.UUID]*api.Actor
	m       sync.Mutex
}

// Get returns a pipeline info
func (pc *ActorCache) Get(id api.UUID) (*api.Actor, error) {
	pc.m.Lock()
	defer pc.m.Unlock()
	if p, ok := pc.cache[id]; ok {
		return p, nil
	}
	rep, err := pc.control.Actor.Query(&control.ActorRequest{})
	if err != nil {
		return nil, err
	}
	for _, p := range rep.GetActorList() {
		pc.cache[p.GetIDAsUUID()] = p
	}
	if p, ok := pc.cache[id]; ok {
		return p, nil
	}
	return nil, ErrNoSuchActor
}
