// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package storage

import (
	"context"
	"time"

	"xbus.io/go-xbus/v4/api"
)

// JobID A job unique id.
type JobID struct {
	ProcessID api.UUID
	Node      ProcessNode
}

// JobState contains all the job state
type JobState struct {
	CreatedAt   time.Time
	Detached    bool
	Status      api.ActorProcessingState_Status
	EnvelopeIDs []api.UUID

	FirstAttempt time.Time
	AttemptCount uint
}

// Job is a director job definition and state
type Job struct {
	JobID
	JobState
}

func (id JobID) String() string {
	return id.ProcessID.String() + "/" + id.Node.String()
}

// JobList is for returning long job lists in a stream
type JobList interface {
	Next() bool
	Get() (Job, error)
	Close()
}

// AllJobs returns a complete list and close the given JobList
func AllJobs(list JobList) ([]Job, error) {
	var jobs []Job
	defer list.Close()
	for list.Next() {
		job, err := list.Get()
		if err != nil {
			return nil, err
		}
		jobs = append(jobs, job)
	}
	return jobs, nil
}

// JobStorage persists the jobs
type JobStorage interface {
	NewJob(context.Context, JobID, JobState)
	SetState(context.Context, JobID, JobState)
	GetState(context.Context, JobID) JobState

	ListPendingJobByActor(ctx context.Context, actor api.UUID, after *JobID, limit int) []Job
	ListJobByProcess(context.Context, api.UUID) []Job
	ListRunningJobByEnvelopeID(context.Context, api.UUID) []Job
	ListDetachedJobByActor(ctx context.Context, actorID api.UUID) (JobList, error)

	Purge(context.Context, []api.UUID) error
}
