// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package metrics

import (
	"context"
	"time"

	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/api/control"

	"xbus.io/xbus/storage"
)

func NewEnvelopeStorage(name, implementation string, backend storage.EnvelopeStorage) *EnvelopeStorage {
	return &EnvelopeStorage{
		timer:   NewStorageTimer("envelope", name, implementation),
		backend: backend,
	}
}

type EnvelopeStorage struct {
	timer   StorageTimer
	backend storage.EnvelopeStorage
}

func (s EnvelopeStorage) SetEventHandler(handler storage.EnvelopeStorageEventHandler) {
	defer s.timer.Time("SetEventHandler")(true)
	s.backend.SetEventHandler(handler)
}

func (s EnvelopeStorage) StoreEnvelope(ctx context.Context, id *api.Envelope) (storage.EnvelopeState, error) {
	observe := s.timer.Time("StoreEnvelope")
	state, err := s.backend.StoreEnvelope(ctx, id)
	observe(err == nil)
	return state, err
}

func (s EnvelopeStorage) CheckStalled(ctx context.Context, id api.UUID, duration time.Duration) bool {
	defer s.timer.Time("CheckStalled")(true)
	return s.backend.CheckStalled(ctx, id, duration)
}

func (s EnvelopeStorage) GetEventTypes(ctx context.Context, id api.UUID) ([]string, bool) {
	defer s.timer.Time("GetEventTypes")(true)
	return s.backend.GetEventTypes(ctx, id)
}

// ReceptionStatus returns the envelope reception status
func (s EnvelopeStorage) ReceptionStatus(ctx context.Context, id api.UUID) (
	storage.EnvelopeStorageReceptionStatus, error,
) {
	observe := s.timer.Time("ReceptionStatus")
	status, err := s.backend.ReceptionStatus(ctx, id)
	observe(err == nil)
	return status, err
}

func (s EnvelopeStorage) ReadEnvelope(
	ctx context.Context, id api.UUID, position storage.EnvelopeReadPos, maxsize int,
) (*api.Envelope, storage.EnvelopeReadPos, error) {
	observe := s.timer.Time("ReadEnvelope")
	e, p, err := s.backend.ReadEnvelope(ctx, id, position, maxsize)
	observe(err == nil)
	return e, p, err
}

func (s EnvelopeStorage) Purge(ctx context.Context, filter storage.EnvelopeFilter) {
	defer s.timer.Time("Purge")(true)
	s.backend.Purge(ctx, filter)
}

func (s EnvelopeStorage) Stat(ctx context.Context) ([]*control.StorageStat_Entry, error) {
	observe := s.timer.Time("Stat")
	stat, err := s.backend.Stat(ctx)
	observe(err == nil)
	return stat, err
}
