// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package sql

import (
	"context"

	"github.com/m4rw3r/uuid"
	"github.com/orus-io/yago"
	"github.com/rs/zerolog"
	"xbus.io/go-xbus/v4/api"
)

// NewPasswordStorage returns a PasswordStorage instance
func NewPasswordStorage(db *yago.DB, logger zerolog.Logger) *PasswordStorage {
	return &PasswordStorage{RawSQLBase{db, logger}}
}

// PasswordStorage provides a SQL based storage.PasswordStorage implementation
type PasswordStorage struct {
	RawSQLBase
}

// Set sets a password the a given id
func (s *PasswordStorage) Set(ctx context.Context, id api.UUID, password string) {
	tx := s.MustBeginContext(ctx)
	defer tx.RollbackIfOpened()

	_, err := tx.ExecContext(ctx,
		"INSERT INTO passwords (account_id, password) "+
			"VALUES ($1, crypt($2, gen_salt('bf'))) "+
			"ON CONFLICT (account_id) "+
			"DO UPDATE SET password = crypt($2, gen_salt('bf'))",
		uuid.UUID(id), password,
	)
	if err != nil {
		panic(err)
	}

	tx.MustCommit()
}

// Reset removed any password for the given id
func (s *PasswordStorage) Reset(ctx context.Context, id api.UUID) {
	tx := s.MustBeginContext(ctx)
	defer tx.RollbackIfOpened()

	_, err := tx.ExecContext(ctx,
		"DELETE FROM passwords "+
			"WHERE account_id = $1",
		uuid.UUID(id),
	)
	if err != nil {
		panic(err)
	}

	tx.MustCommit()
}

// Verify returns true if the given password is correct
func (s *PasswordStorage) Verify(ctx context.Context, id api.UUID, password string) bool {
	tx := s.MustBeginContext(ctx)
	defer func() { _ = tx.Rollback() }()

	row := tx.QueryRowContext(ctx, "SELECT EXISTS("+
		"SELECT 1 "+
		"FROM passwords "+
		"WHERE account_id = $1 "+
		"  AND password = crypt($2, password))",
		uuid.UUID(id), password,
	)

	var correct bool
	if err := row.Scan(&correct); err != nil {
		panic(err)
	}
	return correct
}
