// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package sql

import (
	"context"
	"database/sql"
	"encoding/json"
	"errors"
	"fmt"
	"time"

	"github.com/orus-io/qb"
	"github.com/orus-io/yago"
	"github.com/rs/zerolog"
)

var ErrNoDuration = errors.New("no duration measurement")

// RawSQLBase provides a base type for implementing pure-sql storages
type RawSQLBase struct {
	db     *yago.DB
	logger zerolog.Logger
}

func (sb *RawSQLBase) MustBeginContext(ctx context.Context) *RawSQLTxHelper {
	return RawSQLMustBeginContext(ctx, sb.db, sb.logger)
}

// RawSQLTxHelper wraps a SQL tx and make it easier to enforce commit or rollback
type RawSQLTxHelper struct {
	*sql.Tx
	complete bool
	logflags qb.LogFlags
	logger   zerolog.Logger
}

// RawSQLMustBeginContext starts a sql transaction and wrap it
func RawSQLMustBeginContext(ctx context.Context, db *yago.DB, logger zerolog.Logger) *RawSQLTxHelper {
	tx, err := db.Engine.DB().BeginTx(ctx, nil)
	if err != nil {
		logger.Err(err).Msg("Cannot start a transaction")
		panic(err)
	}
	return &RawSQLTxHelper{tx, false, db.Engine.Logger().LogFlags(), logger}
}

// Commit commits the transaction
func (tx *RawSQLTxHelper) Commit() error {
	err := tx.Tx.Commit()
	tx.complete = true
	if err != nil {
		tx.logger.Err(err).Msg("Error on commit")
	}
	return err
}

// MustCommit commits of rollback
func (tx *RawSQLTxHelper) MustCommit() {
	if err := tx.Commit(); err != nil {
		panic(err)
	}
}

// Rollback rollbacks and log error if any
func (tx *RawSQLTxHelper) Rollback() error {
	err := tx.Tx.Rollback()
	tx.complete = true
	if err != nil {
		tx.logger.Err(err).Msg("Error on rollback")
	}
	return err
}

// BlindRollback rollbacks the transaction, log errors if any, and returns nothing
func (tx *RawSQLTxHelper) BlindRollback() {
	_ = tx.Rollback()
}

// RollbackIfOpened rollback the transaction is still open
func (tx *RawSQLTxHelper) RollbackIfOpened() {
	if !tx.complete {
		_ = tx.Rollback()
	}
}

func finalizeTrace(finalize func(err error), err *error) {
	if r := recover(); r != nil {
		if er, ok := r.(error); ok {
			err = &er
		} else {
			e := fmt.Errorf("caught panic")
			err = &e
		}
		finalize(*err)
		panic(r)
	}
	finalize(*err)
}

func (tx *RawSQLTxHelper) Exec(query string, args ...interface{}) (res sql.Result, err error) {
	defer finalizeTrace(TraceQuery(tx.logflags, &tx.logger, query, args), &err)
	res, err = tx.Tx.Exec(query, args...)
	return
}

func (tx *RawSQLTxHelper) ExecContext(ctx context.Context, query string, args ...interface{}) (res sql.Result, err error) {
	defer finalizeTrace(TraceQuery(tx.logflags, &tx.logger, query, args), &err)
	res, err = tx.Tx.ExecContext(ctx, query, args...)
	return
}

func (tx *RawSQLTxHelper) Query(query string, args ...interface{}) (rows *sql.Rows, err error) {
	defer finalizeTrace(TraceQuery(tx.logflags, &tx.logger, query, args), &err)
	rows, err = tx.Tx.Query(query, args...) //nolint: sqlclosecheck
	return
}

func (tx *RawSQLTxHelper) QueryContext(ctx context.Context, query string, args ...interface{}) (rows *sql.Rows, err error) {
	defer finalizeTrace(TraceQuery(tx.logflags, &tx.logger, query, args), &err)
	rows, err = tx.Tx.QueryContext(ctx, query, args...) //nolint: sqlclosecheck
	return
}

func (tx *RawSQLTxHelper) QueryRow(query string, args ...interface{}) (row *sql.Row) {
	defer finalizeTrace(TraceQuery(tx.logflags, &tx.logger, query, args), &ErrNoDuration)
	row = tx.Tx.QueryRow(query, args...)
	return
}

func (tx *RawSQLTxHelper) QueryRowContext(ctx context.Context, query string, args ...interface{}) (row *sql.Row) {
	defer finalizeTrace(TraceQuery(tx.logflags, &tx.logger, query, args), &ErrNoDuration)
	row = tx.Tx.QueryRowContext(ctx, query, args...)
	return
}

// TraceQuery logs a sql query and args as TRACE level
func TraceQuery(flags qb.LogFlags, log *zerolog.Logger, sql string, args []interface{}) func(error) {
	if log != nil && flags != qb.LDefault && log.GetLevel() <= zerolog.DebugLevel {
		start := time.Now()
		return func(err error) {
			e := log.Debug()
			if flags&qb.LQuery != 0 {
				e = e.Str("query", sql)
			}
			if flags&qb.LBindings != 0 {
				arr := zerolog.Arr()
				for _, arg := range args {
					b, err := json.Marshal(arg)
					if err != nil {
						arr.Interface(arg)
					} else {
						if len(b) > 512 {
							b, _ = json.Marshal(map[string]interface{}{
								"arg first 512 bytes": b[:512],
								"len":                 len(b),
							})
						}
						arr.RawJSON(b)
					}
				}
				e = e.Array("args", arr)
			}
			if err != nil && errors.Is(err, ErrNoDuration) {
				e = e.Bool("no-duration-measure", true)
			} else {
				if err != nil {
					e = e.Err(err)
				}
				if err == nil || !errors.Is(err, ErrNoDuration) {
					e = e.Dur("duration", time.Since(start))
				}
			}
			e.Msg("SQL Query")
		}
	}
	return func(_ error) {}
}
