// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package testutils

import (
	"crypto/tls"
	"crypto/x509"
	"fmt"
	"net"
	"testing"
	"time"

	"github.com/nats-io/nats-server/v2/server"
	nats "github.com/nats-io/nats.go"
	"github.com/rs/zerolog"
	xbus "xbus.io/go-xbus/v4"

	"xbus.io/xbus/auth"
	"xbus.io/xbus/storage"
)

// NatsTestServer wraps a gnatsd server.Server
type NatsTestServer struct {
	srv         *server.Server
	auth        *auth.XBusAuth
	ready       bool
	ConnOptions nats.Options
	cert        *tls.Certificate
}

// InitNatsTestServer returns a gnatsd test server
func InitNatsTestServer(
	t *testing.T, logger zerolog.Logger,
	accountStorage storage.AccountStorage,
	passwordStorage storage.PasswordStorage,
	actorStorage storage.ActorStorage,
	sessionStorage storage.SessionStorage,
	serverCert *tls.Certificate,
) NatsTestServer {
	t.Helper()

	var a *auth.XBusAuth
	if accountStorage != nil {
		aa := auth.NewXBusAuth(
			accountStorage, passwordStorage, actorStorage, sessionStorage, logger)
		a = &aa
	}
	var tlsConfig *tls.Config
	if serverCert != nil {
		certPool := x509.NewCertPool()
		certPool.AddCert(serverCert.Leaf)
		tlsConfig = &tls.Config{
			Certificates: []tls.Certificate{*serverCert},
			ClientAuth:   tls.VerifyClientCertIfGiven,
			ClientCAs:    certPool,
		}
	}

	var customAuth server.Authentication
	if a != nil {
		customAuth = a
	}
	s := NatsTestServer{
		srv: server.New(&server.Options{
			Host: "localhost", Port: 0,
			CustomClientAuthentication: customAuth,
			TLSConfig:                  tlsConfig,
		}),
		auth:  a,
		ready: false,
		cert:  serverCert,
	}
	s.srv.SetLogger(xbus.NewNATSLogger(logger), false, false)

	go s.srv.Start()

	if !s.srv.ReadyForConnections(time.Second) {
		t.Fatalf("Server was not ready for connection after 1 second")
	}

	addr := s.srv.Addr().String()
	conn, err := net.Dial("tcp", addr)
	if err != nil {
		t.Fatalf("Could not connect to the server")
	}
	conn.Close()

	// Wait a bit to give a chance to the server to remove this
	// "client" from its state, which may otherwise interfere with
	// some tests.
	time.Sleep(5 * time.Millisecond)
	s.ready = true

	var connTLSConfig *tls.Config
	if tlsConfig != nil {
		connTLSConfig = &tls.Config{
			RootCAs:    tlsConfig.ClientCAs,
			ServerName: "localhost",
		}
	}
	s.ConnOptions = nats.Options{
		Url:       fmt.Sprintf("nats://%s", s.srv.Addr().String()),
		Secure:    connTLSConfig != nil,
		TLSConfig: connTLSConfig,
		AsyncErrorCB: func(c *nats.Conn, s *nats.Subscription, err error) {
			if s != nil {
				logger.Error().Msgf("Error on sub %s: %s", s.Subject, err)
			} else {
				logger.Error().Msgf("Error %s", err)
			}
		},
	}
	if !s.ready {
		t.Fatalf("Unable to start the server properly (cannot connect to it)")
	}

	return s
}

// IsReady returns true if the server is ready
func (s NatsTestServer) IsReady() bool {
	return s.ready
}

// Connect returns a new client connection to the server
func (s NatsTestServer) Connect(t *testing.T) *nats.Conn { // Tests are all done on localhost
	t.Helper()

	opts := s.ConnOptions
	if s.auth != nil {
		opts.Token = s.auth.AuthToken
	}
	natsConn, err := opts.Connect()

	if err != nil {
		t.Errorf("Cannot connect to nats: %s", err)
		return nil
	}
	return natsConn
}

// Shutdown stops the server
func (s NatsTestServer) Shutdown() {
	s.srv.Shutdown()
}

// AuthToken returns the auth token
func (s NatsTestServer) AuthToken() string {
	return s.auth.AuthToken
}
