// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package storage

import (
	"testing"

	"xbus.io/go-xbus/v4/api"

	"xbus.io/xbus/storage"
	"xbus.io/xbus/testutils"
)

const purgeBatchSize = 5000

type massCreateProcessFunc func(tb testing.TB, ps storage.ProcessStorage, count int, status api.Process_Status) []api.UUID

// ProcessStorageBenchmark is a benchmark suite for a ProcessStorage implementation
func ProcessStorageBenchmark(
	b *testing.B, factory ProcessStorageFactory, massCreateProcess massCreateProcessFunc,
) {
	b.Helper()
	ctx := testutils.GetTestContext(b)

	if massCreateProcess == nil {
		massCreateProcess = func(tb testing.TB, ps storage.ProcessStorage, count int, status api.Process_Status) []api.UUID {
			tb.Helper()

			ids := make([]api.UUID, 0, count)
			for i := 0; i != count; i++ {
				id, err := ps.CreateProcess(ctx, storage.Process{})
				if err != nil {
					b.Fatal(err)
				}
				if err := ps.SetStatus(ctx, id, status, ""); err != nil {
					b.Fatal(err)
				}
				ids = append(ids, id)
			}
			return ids
		}
	}
	b.Run("Purge", func(b *testing.B) {
		ps, clean := factory(b)
		defer clean()
		ids := massCreateProcess(b, ps, b.N, api.Process_DONE)

		b.ResetTimer()
		for i := 0; i < b.N/purgeBatchSize; i++ {
			if res, err := ps.PurgeProcess(ctx, ids[i*purgeBatchSize:i*purgeBatchSize+purgeBatchSize]); err != nil {
				b.Fatal(b)
			} else if len(res) != purgeBatchSize {
				b.Fatalf("Expected %d process purged, got: %d", purgeBatchSize, len(res))
			}
		}
		if res, err := ps.PurgeProcess(ctx, ids[b.N/purgeBatchSize*purgeBatchSize:]); err != nil {
			b.Fatal(b)
		} else if len(res) != b.N%purgeBatchSize {
			b.Fatalf("Expected %d process purged, got: %d", b.N%purgeBatchSize, len(res))
		}
	})
}
