// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package storage

import (
	"testing"
	"time"

	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	"google.golang.org/protobuf/types/known/timestamppb"
	"xbus.io/go-xbus/v4/api"

	"xbus.io/xbus/storage"
	"xbus.io/xbus/testutils"
)

// SessionStorageFactory is any factory of SessionStorage
type SessionStorageFactory func(bt testing.TB) (storage.SessionStorage, func())

// SessionStorageTests is a test suite for SessionStorage implementations
func SessionStorageTests(t *testing.T, factory SessionStorageFactory) {
	t.Helper()

	t.Run("SetAndGet", func(t *testing.T) {
		ctx := testutils.GetTestContext(t)
		store, clean := factory(t)
		if store == nil {
			t.Fatal("No store returned by the factory")
		}
		defer clean()

		session := storage.Session{
			SessionToken: &api.SessionToken{
				Token:      "thetoken",
				ValidUntil: timestamppb.New(time.Date(2019, 8, 27, 20, 0, 0, 0, time.UTC)),
			},
			AccountID: api.MustUUIDFromString("483b8532-a6f2-11e8-b206-13d021cd09cd"),
		}

		err := store.Set(ctx, session)
		if err != nil {
			t.Fatal(err)
		}

		s, err := store.Get(ctx, session.Token)
		if err != nil {
			t.Fatal(err)
		}

		if s == nil {
			t.Fatal("Get() returned (nil, nil)")
		}
		assert.Equal(t, session, *s)
	})

	t.Run("List", func(t *testing.T) {
		ctx := testutils.GetTestContext(t)
		store, clean := factory(t)
		if store == nil {
			t.Fatal("No store returned by the factory")
		}
		defer clean()

		list, err := store.List(ctx)
		if err != nil {
			t.Fatal(err)
		}

		assert.Empty(t, list)

		session := storage.Session{
			SessionToken: &api.SessionToken{
				Token:      "thetoken",
				ValidUntil: timestamppb.New(time.Date(2019, 8, 27, 20, 0, 0, 0, time.UTC)),
			},
			AccountID: api.MustUUIDFromString("483b8532-a6f2-11e8-b206-13d021cd09cd"),
		}

		if err := store.Set(ctx, session); err != nil {
			t.Fatal(err)
		}

		list, err = store.List(ctx)
		if err != nil {
			t.Fatal(err)
		}

		assert.Equal(t, []storage.Session{session}, list)
	})

	t.Run("Delete", func(t *testing.T) {
		ctx := testutils.GetTestContext(t)
		store, clean := factory(t)
		if store == nil {
			t.Fatal("No store returned by the factory")
		}
		defer clean()

		session := storage.Session{
			SessionToken: &api.SessionToken{
				Token:      "thetoken",
				ValidUntil: timestamppb.New(time.Date(2019, 8, 27, 20, 0, 0, 0, time.UTC)),
			},
			AccountID: api.MustUUIDFromString("483b8532-a6f2-11e8-b206-13d021cd09cd"),
		}

		if err := store.Delete(ctx, session.Token); err != nil {
			t.Fatal(err)
		}

		if err := store.Set(ctx, session); err != nil {
			t.Fatal(err)
		}

		s, err := store.Get(ctx, session.Token)
		if err != nil {
			t.Fatal(err)
		}

		if s == nil {
			t.Fatal("Get() returned (nil, nil)")
		}
		assert.Equal(t, session, *s)

		if err := store.Delete(ctx, session.Token); err != nil {
			t.Fatal(err)
		}

		s, err = store.Get(ctx, session.Token)
		require.ErrorIs(t, err, storage.ErrNotExist)
		assert.Nil(t, s)

		list, err := store.List(ctx)
		if err != nil {
			t.Fatal(err)
		}
		assert.Empty(t, list)
	})

	t.Run("DeleteOlderThan", func(t *testing.T) {
		ctx := testutils.GetTestContext(t)
		store, clean := factory(t)
		if store == nil {
			t.Fatal("No store returned by the factory")
		}
		defer clean()

		session1 := storage.Session{
			SessionToken: &api.SessionToken{
				Token:      "thetoken1",
				ValidUntil: timestamppb.New(time.Date(2019, 8, 27, 20, 0, 0, 0, time.UTC)),
			},
			AccountID: api.MustUUIDFromString("483b8532-a6f2-11e8-b206-13d021cd09cd"),
		}

		session2 := storage.Session{
			SessionToken: &api.SessionToken{
				Token:      "thetoken2",
				ValidUntil: timestamppb.New(time.Date(2019, 8, 27, 21, 0, 0, 0, time.UTC)),
			},
			AccountID: api.MustUUIDFromString("483b8532-a6f2-11e8-b206-13d021cd09cd"),
		}

		if err := store.Set(ctx, session1); err != nil {
			t.Fatal(err)
		}

		if err := store.Set(ctx, session2); err != nil {
			t.Fatal(err)
		}

		list, err := store.List(ctx)
		if err != nil {
			t.Fatal(err)
		}

		assert.Equal(t, 2, len(list))
		assert.Contains(t, list, session1)
		assert.Contains(t, list, session2)

		if err := store.DeleteOlderThan(
			ctx, time.Date(2019, 8, 27, 20, 30, 0, 0, time.UTC),
		); err != nil {
			t.Fatal(err)
		}

		list, err = store.List(ctx)
		if err != nil {
			t.Fatal(err)
		}

		assert.Equal(t, 1, len(list))
		assert.Contains(t, list, session2)
	})
}
