package api

import (
	"bytes"
	"encoding/base64"
	"encoding/json"
	"fmt"
	"time"
	"unicode/utf8"

	jsoniter "github.com/json-iterator/go"
	"google.golang.org/protobuf/types/known/timestamppb"
	"xbus.io/go-xbus/v4/utils"
)

func GetJSONMarshalContext(stream *jsoniter.Stream) *JSONMarshalContext {
	if c, ok := stream.Attachment.(*JSONMarshalContext); ok {
		return c
	}
	return nil
}

type JSONMarshalContext struct {
	acceptEncodings []ItemEncoding
}

func (c *JSONMarshalContext) AcceptEncodings() []ItemEncoding {
	if c == nil {
		return []ItemEncoding{ItemJSON, ItemText, ItemBinary}
	}
	return c.acceptEncodings
}

func jsonMarshalGetAcceptEncoding(stream *jsoniter.Stream) []ItemEncoding {
	return GetJSONMarshalContext(stream).AcceptEncodings()
}

func JSONStreamWriteUUID(stream *jsoniter.Stream, id []byte) {
	stream.WriteString(MustUUIDFromBytes(id).String())
}

func JSONStreamWriteTimestamp(stream *jsoniter.Stream, t *timestamppb.Timestamp) {
	stream.WriteString(t.AsTime().Format(time.RFC3339))
}

func JSONStreamItemWriter(encoding ItemEncoding) func(stream *jsoniter.Stream, item []byte) {
	switch encoding {
	case ItemBinary:
		return func(stream *jsoniter.Stream, item []byte) {
			stream.WriteString(base64.StdEncoding.EncodeToString(item))
		}
	case ItemText:
		return func(stream *jsoniter.Stream, item []byte) {
			// Check if it is valid utf-8
			if !utf8.Valid(item) {
				stream.Error = fmt.Errorf("Not a valid UTF-8 string")
			} else {
				stream.WriteString(string(item))
			}
		}
	case ItemJSON:
		return func(stream *jsoniter.Stream, item []byte) {
			if err := json.Compact(bytes.NewBuffer(nil), item); err != nil {
				stream.Error = fmt.Errorf("Not a valid JSON item")
			} else {
				stream.WriteRaw(string(item))
			}
		}
	default:
		panic(fmt.Sprintf("invalid ItemEncoding: %s", encoding))
	}
}

// MarshalJSONStream encode a Event with an additional "encoding" that is
// the actual encoding of the items
func (x *Event) MarshalJSONStream(stream *jsoniter.Stream) {
	subStream := jsoniter.ConfigDefault.BorrowStream(nil)
	defer jsoniter.ConfigDefault.ReturnStream(subStream)

	obj := utils.NewJSONStreamObject(stream)
	if !x.GetIDAsUUID().IsZero() {
		obj.Field("id")
		JSONStreamWriteUUID(stream, x.ID)
	}

	if x.Type != "" {
		obj.String("type", x.Type)
	}

	if x.Index != 0 {
		obj.Uint64("index", x.Index)
	}

	if x.ItemCount != 0 {
		obj.Uint64("item_count", x.ItemCount)
	}

	if x.Checksum != 0 {
		obj.Uint32("checksum", x.Checksum)
	}

	obj.Field("encoding")

	buflen := len(stream.Buffer())

encodingLoop:
	for _, encoding := range jsonMarshalGetAcceptEncoding(stream) {
		b, _ := encoding.MarshalText()
		stream.WriteString(string(b))

		array := obj.Array("items")

		itemWriter := JSONStreamItemWriter(encoding)
		for _, item := range x.Items {
			array.Add()
			itemWriter(stream, item)
			if stream.Error != nil {
				array.End() // make sure indentation is set back property
				stream.Error = nil
				stream.SetBuffer(stream.Buffer()[:buflen])
				continue encodingLoop
			}
		}
		array.End()

		obj.End()

		return
	}

	stream.Error = fmt.Errorf(
		"No suitable encoding for event %s in %v",
		x.GetIDAsUUID(), jsonMarshalGetAcceptEncoding(stream))
}

func (x *Envelope) MarshalJSONStream(stream *jsoniter.Stream) {
	obj := utils.NewJSONStreamObject(stream)

	if !x.GetIDAsUUID().IsZero() {
		obj.Field("id")
		JSONStreamWriteUUID(stream, x.ID)
	}

	if x.Last {
		obj.True("last")
	}

	a := obj.Array("event_ids")
	for _, id := range x.EventIDs {
		a.Add()
		JSONStreamWriteUUID(stream, id)
	}
	a.End()

	obj.Field("events")
	utils.JSONStreamWriteArray(stream, x.Events)

	obj.End()
}

func (x *EventPosition) MarshalJSONStream(stream *jsoniter.Stream) {
	obj := utils.NewJSONStreamObject(stream)

	if !x.GetEventIDAsUUID().IsZero() {
		obj.Field("event_id")
		JSONStreamWriteUUID(stream, x.EventID)
	}
	obj.Uint64("index", x.Index)
	obj.Uint64("item_count", x.ItemCount)
	obj.End()
}

func (x *EnvelopePosition) MarshalJSONStream(stream *jsoniter.Stream) {
	obj := utils.NewJSONStreamObject(stream)

	if !x.GetEnvelopeIDAsUUID().IsZero() {
		obj.Field("envelope_id")
		JSONStreamWriteUUID(stream, x.EnvelopeID)
	}

	if x.Start {
		obj.True("start")
	}

	if x.Complete {
		obj.True("complete")
	}

	obj.End()
}

func (x *ProcessingContext) MarshalJSONStream(stream *jsoniter.Stream) {
	obj := utils.NewJSONStreamObject(stream)
	if !x.GetProcessIDAsUUID().IsZero() {
		obj.Field("process_id")
		JSONStreamWriteUUID(stream, x.ProcessID)
	}
	if x.NodeID != "" {
		obj.String("node_id", x.NodeID)
	}

	obj.End()
}

func (x *ActorProcessRequest_Input) MarshalJSONStream(stream *jsoniter.Stream) {
	obj := utils.NewJSONStreamObject(stream)

	obj.String("name", x.Name)

	obj.Bool("close", x.Close)

	if x.Envelope != nil {
		obj.Val("envelope", x.Envelope)
	}

	if x.Position != nil {
		obj.Val("position", x.Position)
	}

	obj.End()
}

func (x *ActorProcessRequest) MarshalJSONStream(stream *jsoniter.Stream) {
	obj := utils.NewJSONStreamObject(stream)

	obj.Val("context", x.Context)

	obj.Field("inputs")
	utils.JSONStreamWriteArray(stream, x.Inputs)

	obj.End()
}

func (x *LogMessage) MarshalJSONStream(stream *jsoniter.Stream) {
	obj := utils.NewJSONStreamObject(stream)

	if x.Time != nil {
		obj.Field("time")
		JSONStreamWriteTimestamp(stream, x.Time)
	}
	if x.Level != 0 {
		obj.String("level", x.Level.String())
	}
	if x.Text != "" {
		obj.String("text", x.Text)
	}
	obj.End()
}

func (x *ActorProcessingState) MarshalJSONStream(stream *jsoniter.Stream) {
	obj := utils.NewJSONStreamObject(stream)

	if x.Context != nil {
		obj.Val("context", x.Context)
	}

	obj.String("status", x.Status.String())

	if x.ActorLeaving {
		obj.True("actor_leaving")
	}

	if len(x.Messages) != 0 {
		obj.Field("messages")
		utils.JSONStreamWriteArray(stream, x.Messages)
	}

	obj.End()
}

func (x *LogEntry) MarshalJSONStream(stream *jsoniter.Stream) {
	obj := utils.NewJSONStreamObject(stream)
	if id := x.GetEnvelopeIDAsUUID(); !id.IsZero() {
		obj.Val("envelope_id", id)
	}
	if id := x.GetActorIDAsUUID(); !id.IsZero() {
		obj.Val("actor_id", id)
	}
	if id := x.GetProcessIDAsUUID(); !id.IsZero() {
		obj.Val("process_id", id)
	}
	obj.OEString("node_id", x.NodeID)
	if x.Message != nil {
		obj.Val("message", x.Message)
	}
	defer obj.End()
}

// JSONMarshal encodes api messages to JSON. The encodings list controls how
// the api.Item will be encoded
func JSONMarshal(encodings []ItemEncoding, value utils.JSONStreamMarshaler) ([]byte, error) {
	stream := jsoniter.ConfigDefault.BorrowStream(nil)
	defer jsoniter.ConfigDefault.ReturnStream(stream)

	if err := JSONMarshalToStream(encodings, value, stream); err != nil {
		return nil, err
	}

	return stream.Buffer(), nil
}

// JSONMarshalToStream encodes api messages to JSON. The encodings list controls how
// the api.Item will be encoded
func JSONMarshalToStream(
	encodings []ItemEncoding, value utils.JSONStreamMarshaler, stream *jsoniter.Stream,
) error {
	if len(encodings) == 0 {
		encodings = []ItemEncoding{ItemJSON, ItemText, ItemBinary}
	}
	mctx := JSONMarshalContext{
		acceptEncodings: encodings,
	}

	savedAttachement := stream.Attachment
	defer func() {
		stream.Attachment = savedAttachement
	}()

	stream.Attachment = &mctx
	value.MarshalJSONStream(stream)

	if stream.Error != nil {
		return stream.Error
	}

	return stream.Flush()
}
