package api

import (
	"encoding/base64"
	"fmt"
	"io"
	"strings"
	"time"

	jsoniter "github.com/json-iterator/go"
	"google.golang.org/protobuf/types/known/timestamppb"
	"xbus.io/go-xbus/v4/utils"
)

/* Custom json marshaler/unmarshaler for Envelope and all the types that uses
it (OutputRequest, ActorProcessRequest) */

func DecodeJSON[T utils.JSONIterUnmarshaler](r io.Reader, value T) error {
	return utils.DecodeJSON(r, value)
}

func UnmarshalJSON[T utils.JSONIterUnmarshaler](data []byte, value T) error {
	return utils.UnmarshalJSON(data, value)
}

//ItemEncoding The idem encoding in a json payload
//go:generate enum-marshaler -type ItemEncoding -prefix Item -lower -tests
type ItemEncoding int

const (
	// ItemBinary is for base64 encoded data
	ItemBinary ItemEncoding = iota
	// ItemText is for text encoded data
	ItemText
	// ItemJSON is for embedding json directly
	ItemJSON
)

// ParseItemEncodingList parses a comma separated list of item encodings
// Space characters are trimmed from each item
func ParseItemEncodingList(s string) ([]ItemEncoding, error) {
	var itemEncodings []ItemEncoding
	for _, encoding := range strings.Split(s, ",") {
		encoding = strings.Trim(encoding, " ")
		if encoding == "" {
			continue
		}
		var item ItemEncoding
		if err := item.UnmarshalText([]byte(encoding)); err != nil {
			return nil, err
		}
		itemEncodings = append(itemEncodings, item)
	}
	return itemEncodings, nil
}

func JSONIterReadItemEncoding(iter *jsoniter.Iterator) ItemEncoding {
	var res ItemEncoding
	s := iter.ReadString()
	if iter.Error == nil {
		if err := res.UnmarshalText([]byte(s)); err != nil {
			iter.ReportError("parse encoding", err.Error())
		}
	}
	return res
}

func JSONIterItemReader(encoding ItemEncoding) (func(*jsoniter.Iterator) Item, error) {
	switch encoding {
	case ItemBinary:
		return func(iter *jsoniter.Iterator) Item {
			s := iter.ReadString()
			if iter.Error != nil {
				return nil
			}
			b, err := base64.StdEncoding.DecodeString(s)
			if err != nil {
				iter.ReportError("decode item", fmt.Sprintf("could not decode base64 item: %s", err))
				return nil
			}
			return Item(b)
		}, nil
	case ItemText:
		return func(iter *jsoniter.Iterator) Item {
			s := iter.ReadString()
			if iter.Error != nil {
				return nil
			}
			return Item(s)
		}, nil
	case ItemJSON:
		return func(iter *jsoniter.Iterator) Item {
			b := iter.SkipAndReturnBytes()
			if iter.Error != nil {
				return nil
			}
			return Item(b)
		}, nil
	default:
		return nil, fmt.Errorf("Unsupported encoding: %s", encoding)
	}
}

func JSONIterReadItems(iter *jsoniter.Iterator, encoding ItemEncoding) [][]byte {
	reader, err := JSONIterItemReader(encoding)
	if err != nil {
		iter.ReportError("decode items", err.Error())
		return nil
	}
	var (
		itemList [][]byte
	)
	if iter.ReadArrayCB(func(i *jsoniter.Iterator) bool {
		item := reader(i)
		if i.Error != nil {
			return false
		}
		itemList = append(itemList, item)
		return true
	}) {
		return itemList
	}
	return nil
}

func JSONIterReadUUID(iter *jsoniter.Iterator) UUID {
	s := iter.ReadString()
	if iter.Error != nil {
		return UUID{}
	}
	id, err := UUIDFromString(s)
	if err != nil {
		iter.ReportError("decode UUID", err.Error())
	}
	return id
}

func JSONIterReadTimestamp(iter *jsoniter.Iterator) *timestamppb.Timestamp {
	value := iter.ReadString()
	if iter.Error != nil {
		return nil
	}
	t, err := time.Parse(time.RFC3339, value)
	if err != nil {
		iter.ReportError("parse time", fmt.Sprintf("could not parse the time: %s.", err))
		return nil
	}
	return timestamppb.New(t)
}

func (x *Event) UnmarshalJSONIter(iter *jsoniter.Iterator) {
	if x == nil {
		panic("cannot unmarshall to nil *Event")
	}
	var (
		encoding *ItemEncoding
		items    []byte
	)
	if iter.ReadObjectCB(func(i *jsoniter.Iterator, s string) bool {
		switch s {
		case "id", "Id", "ID":
			x.ID = JSONIterReadUUID(i).ToBytes()
		case "type", "Type":
			x.Type = i.ReadString()
		case "index", "Index":
			x.Index = i.ReadUint64()
		case "item_count", "ItemCount":
			x.ItemCount = i.ReadUint64()
		case "checksum", "Checksum":
			x.Checksum = i.ReadUint32()
		case "encoding", "Encoding":
			e := JSONIterReadItemEncoding(i)
			if i.Error == nil {
				encoding = &e
			}
			if items != nil {
				subIter := jsoniter.ConfigDefault.BorrowIterator(items)
				defer jsoniter.ConfigDefault.ReturnIterator(subIter)

				x.Items = JSONIterReadItems(subIter, *encoding)
				if subIter.Error != nil {
					iter.ReportError("parsing items", subIter.Error.Error())
				}
			}
		case "items", "Items":
			if encoding != nil {
				x.Items = JSONIterReadItems(iter, *encoding)
			} else {
				items = iter.SkipAndReturnBytes()
			}
		}
		return i.Error == nil
	}) {
		if encoding == nil && len(items) != 0 {
			subIter := jsoniter.ConfigDefault.BorrowIterator(items)
			defer jsoniter.ConfigDefault.ReturnIterator(subIter)

			x.Items = JSONIterReadItems(subIter, ItemBinary)
			if subIter.Error != nil {
				iter.ReportError("parsing items", subIter.Error.Error())
			}
		}
	}
}

func (x *Envelope) UnmarshalJSONIter(iter *jsoniter.Iterator) {
	iter.ReadObjectCB(func(i *jsoniter.Iterator, s string) bool {
		switch s {
		case "id", "Id", "ID":
			x.ID = JSONIterReadUUID(i).ToBytes()
		case "event_ids", "EventIDs":
			iter.ReadArrayCB(func(i *jsoniter.Iterator) bool {
				x.EventIDs = append(x.EventIDs, JSONIterReadUUID(i).ToBytes())
				return i.Error == nil
			})
		case "last", "Last":
			x.Last = i.ReadBool()
		case "events", "Events":
			iter.ReadArrayCB(func(i *jsoniter.Iterator) bool {
				x.Events = append(x.Events, utils.JSONIterReadPointer[*Event](i))
				return i.Error == nil
			})
		}
		return i.Error == nil
	})
}

func (x *ProcessingContext) UnmarshalJSONIter(iter *jsoniter.Iterator) {
	iter.ReadObjectCB(func(i *jsoniter.Iterator, s string) bool {
		switch s {
		case "process_id", "processID", "ProcessID":
			x.ProcessID = JSONIterReadUUID(i).ToBytes()
		case "node_id", "nodeID", "NodeID":
			x.NodeID = i.ReadString()
		}
		return i.Error == nil
	})
}

func (x *EventPosition) UnmarshalJSONIter(iter *jsoniter.Iterator) {
	iter.ReadObjectCB(func(i *jsoniter.Iterator, s string) bool {
		switch s {
		case "event_id", "eventID", "EventID":
			x.EventID = JSONIterReadUUID(i).ToBytes()
		case "index", "Index":
			x.Index = i.ReadUint64()
		case "item_count", "itemCount", "ItemCount":
			x.ItemCount = i.ReadUint64()
		}
		return i.Error == nil
	})
}

func (x *EnvelopePosition) UnmarshalJSONIter(iter *jsoniter.Iterator) {
	iter.ReadObjectCB(func(i *jsoniter.Iterator, s string) bool {
		switch s {
		case "envelope_id", "envelopeID", "EnvelopeID":
			x.EnvelopeID = JSONIterReadUUID(i).ToBytes()
		case "start", "Start":
			x.Start = i.ReadBool()
		case "complete", "Complete":
			x.Complete = i.ReadBool()
		case "event_positions", "eventPositions", "EventPositions":
			i.ReadArrayCB(func(i *jsoniter.Iterator) bool {
				x.EventPositions = append(x.EventPositions, utils.JSONIterReadPointer[*EventPosition](i))
				return i.Error == nil
			})
		}
		return i.Error == nil
	})
}

func (x *ActorProcessRequest_Input) UnmarshalJSONIter(iter *jsoniter.Iterator) {
	iter.ReadObjectCB(func(i *jsoniter.Iterator, s string) bool {
		switch s {
		case "name", "Name":
			x.Name = i.ReadString()
		case "close", "Close":
			x.Close = i.ReadBool()
		case "envelope", "Envelope":
			var e Envelope
			e.UnmarshalJSONIter(i)
			if i.Error == nil {
				x.Envelope = &e
			}
		case "position", "Position":
			x.Position = utils.JSONIterReadPointer[*EnvelopePosition](i)
		}
		return i.Error == nil
	})
}

func (x *ActorProcessRequest) UnmarshalJSONIter(iter *jsoniter.Iterator) {
	iter.ReadObjectCB(func(i *jsoniter.Iterator, s string) bool {
		switch s {
		case "context", "Context":
			x.Context = utils.JSONIterReadPointer[*ProcessingContext](i)
		case "inputs", "Inputs":
			i.ReadArrayCB(func(i *jsoniter.Iterator) bool {
				x.Inputs = append(x.Inputs, utils.JSONIterReadPointer[*ActorProcessRequest_Input](i))
				return i.Error == nil
			})
		}
		return i.Error == nil
	})
}

func (x *OutputRequest) UnmarshalJSONIter(iter *jsoniter.Iterator) {
	iter.ReadObjectCB(func(i *jsoniter.Iterator, s string) bool {
		switch s {
		case "context", "Context":
			x.Context = utils.JSONIterReadPointer[*ProcessingContext](i)
		case "output", "Output":
			x.Output = i.ReadString()
		case "close", "Close":
			x.Close = i.ReadBool()
		case "envelope", "Envelope":
			x.Envelope = utils.JSONIterReadPointer[*Envelope](i)
		}
		return i.Error == nil
	})
}

func (x *LogMessage) UnmarshalJSONIter(iter *jsoniter.Iterator) {
	iter.ReadObjectCB(func(i *jsoniter.Iterator, s string) bool {
		switch s {
		case "time", "Time":
			x.Time = JSONIterReadTimestamp(i)
		case "level", "Level":
			x.Level = LogLevel(LogLevel_value[i.ReadString()])
		case "text", "Text":
			x.Text = i.ReadString()
		}
		return i.Error == nil
	})
}

func (x *ActorProcessingState) UnmarshalJSONIter(iter *jsoniter.Iterator) {
	iter.ReadObjectCB(func(i *jsoniter.Iterator, s string) bool {
		switch s {
		case "context", "Context":
			x.Context = utils.JSONIterReadPointer[*ProcessingContext](i)
		case "status", "Status":
			x.Status = ActorProcessingState_Status(
				ActorProcessingState_Status_value[i.ReadString()])
		case "messages", "Messages":
			x.Messages = utils.JSONIterReadArray[*LogMessage](i)
		case "actor_leaving", "actorLeaving", "ActorLeaving":
			x.ActorLeaving = i.ReadBool()
		}
		return i.Error == nil
	})
}
