package cmd

import (
	"context"
	"fmt"
	"io"
	"os"

	jsoniter "github.com/json-iterator/go"
	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/envelope"
)

func loadEnvelope(input string) (envelope.Envelope, error) {
	var in io.Reader
	if input == "-" {
		in = os.Stdin
	} else {
		f, err := os.Open(input)
		if err != nil {
			return nil, err
		}
		defer f.Close()
		in = f
	}
	var fragment api.Envelope
	if err := api.DecodeJSON(in, &fragment); err != nil {
		return nil, err
	}
	envelopeID := fragment.GetIDAsUUID()
	if envelopeID.IsZero() {
		envelopeID = mustUUIDV4()
	}
	builder := envelope.NewOpenedEnvelopeWithID(envelopeID)
	for _, event := range fragment.Events {
		eventID := event.GetIDAsUUID()
		if eventID.IsZero() {
			eventID = mustUUIDV4()
		}
		writer := envelope.NewMessageWriterWithID(eventID, event.Type)
		if len(event.Items) == 1 {
			if _, err := writer.Write(event.Items[0]); err != nil {
				return nil, err
			}
		} else {
			for _, item := range event.Items {
				if err := writer.WriteChunk(item); err != nil {
					return nil, err
				}
			}
		}
		if err := writer.Close(); err != nil {
			return nil, err
		}
		if err := builder.Add(writer); err != nil {
			return nil, err
		}
	}
	if err := builder.Close(); err != nil {
		return nil, err
	}
	return builder, nil
}

var indentedJsonCfg = jsoniter.Config{
	EscapeHTML:    true,
	IndentionStep: 2,
}.Froze()

func saveEnvelope(ctx context.Context, env envelope.Envelope, output string, indent bool) (retErr error) {
	var out io.Writer
	if output == "-" {
		out = os.Stdout
	} else {
		f, err := os.Create(output)
		if err != nil {
			return err
		}
		out = f
		defer func() {
			err := f.Close()
			if err != nil {
				if retErr != nil {
					retErr = fmt.Errorf("%w - could not close file: %w", retErr, err)
				} else {
					retErr = nil
				}
			}
		}()
	}
	complete, err := envelope.CompleteWithContext(ctx, env)
	if err != nil {
		return err
	}
	jsonCfg := jsoniter.ConfigDefault
	if indent {
		jsonCfg = indentedJsonCfg
	}
	stream := jsonCfg.BorrowStream(out)
	defer jsonCfg.ReturnStream(stream)

	if err := api.JSONMarshalToStream(
		[]api.ItemEncoding{api.ItemJSON, api.ItemText, api.ItemBinary},
		complete,
		stream,
	); err != nil {
		return err
	}
	return
}
