package envelope

import (
	"fmt"
	"io"
	"sync"

	"xbus.io/go-xbus/v4/api"
)

// NewStreamMessage creates a StreamMessage
func NewStreamMessage(id api.UUID, msgType string) *StreamMessage {
	m := StreamMessage{
		id:          id,
		msgType:     msgType,
		curChecksum: api.Checksum(0).Update(id[:16]).Update([]byte(msgType)),
	}
	m.cond = sync.NewCond(&m.mut)
	return &m
}

// StreamMessage is a streamed message
type StreamMessage struct {
	id          api.UUID
	msgType     string
	checksum    api.Checksum
	curChecksum api.Checksum

	mut    sync.Mutex
	cond   *sync.Cond
	chunks [][]byte
	err    error
	done   bool

	clones []*StreamMessage
}

type streamMessageChunkReader struct {
	m *StreamMessage
}

func (r *streamMessageChunkReader) DataAvailable() bool {
	r.m.mut.Lock()
	defer r.m.mut.Unlock()

	return r.m.done || r.m.err == nil && len(r.m.chunks) != 0
}

func (r *streamMessageChunkReader) ReadBytes() ([]byte, error) {
	r.m.mut.Lock()

	defer r.m.mut.Unlock()

	for r.m.err == nil && !r.m.done && len(r.m.chunks) == 0 {
		r.m.cond.Wait()
	}
	if r.m.err != nil {
		return nil, r.m.err
	} else if len(r.m.chunks) == 0 {
		if !r.m.done {
			panic("invalid state")
		}
		return nil, io.EOF
	} else {
		chunk := r.m.chunks[0]
		r.m.chunks = r.m.chunks[1:]
		return chunk, nil
	}
}

func (r *streamMessageChunkReader) ReadString() (string, error) {
	b, err := r.ReadBytes()
	if err != nil {
		return "", err
	}
	return string(b), err
}

// NewChunkReaderReader returns a io.Reader that consumes a ChunkReader
func NewChunkReaderReader(chunkReader ChunkReader) io.Reader {
	return &ChunkReaderReader{
		chunkreader: chunkReader,
		buf:         make([]byte, MaxChunkSize),
	}
}

// ChunkReaderReader implements io.Reader on top of ChunkReader
type ChunkReaderReader struct {
	chunkreader ChunkReader

	buf      []byte
	curChunk []byte
}

// Read  implements the io.Reader interface
func (r *ChunkReaderReader) Read(buf []byte) (int, error) {
	// TODO if buf is bigger than curChunk and we have data available,
	// read as much data as we can
	if len(r.curChunk) == 0 {
		chunk, err := r.chunkreader.ReadBytes()
		if err != nil {
			return 0, err
		}
		r.curChunk = chunk
	}
	n := copy(buf, r.curChunk)
	r.curChunk = r.curChunk[n:]
	return n, nil
}

// CloseWithError set a the given error on the message, which will block any
// subsequent read/write
func (m *StreamMessage) CloseWithError(err error) error {
	m.mut.Lock()
	if m.err != nil {
		err := m.err
		m.mut.Unlock()
		return err
	}
	m.err = err
	m.mut.Unlock()
	m.cond.Broadcast()
	return nil
}

func (m *StreamMessage) addChunks(chunks [][]byte) error {
	m.mut.Lock()
	// if append panics we don't want a deadlock in addition
	defer m.mut.Unlock()

	if m.err != nil {
		return m.err
	}

	if chunks == nil {
		if m.checksum == 0 {
			m.checksum = m.curChecksum
		} else {
			if m.curChecksum != m.checksum {
				err := fmt.Errorf("%w: expected 0x%x, got 0x%x", ErrCheckum, m.checksum, m.curChecksum)
				m.err = err
				m.cond.Broadcast()
				return err
			}
		}
		m.done = true
	} else {
		m.chunks = append(m.chunks, chunks...)
		for _, chunk := range chunks {
			m.curChecksum = m.curChecksum.Update(chunk)
		}
	}

	m.cond.Broadcast()

	for _, clone := range m.clones {
		// XXX is it alright to ignore clone errors ?
		_ = clone.addChunks(chunks)
	}
	return nil
}

// ID returns the message ID
func (m *StreamMessage) ID() api.UUID {
	return m.id
}

// MsgType returns the message type
func (m *StreamMessage) MsgType() string {
	return m.msgType
}

// Checksum returns the message checksum if known, 0 otherwise
func (m *StreamMessage) Checksum() (c api.Checksum) {
	m.mut.Lock()
	c = m.checksum
	m.mut.Unlock()
	return
}

// SetChecksum sets the checksum. If called, the calculated checksum will be
// compared to it on closing
func (m *StreamMessage) SetChecksum(c api.Checksum) {
	m.mut.Lock()
	m.checksum = c
	m.mut.Unlock()
}

// Clone the message. Cannot be called if any content access method was
// already called
func (m *StreamMessage) Clone() Message {
	m.mut.Lock()
	defer m.mut.Unlock()

	clone := NewStreamMessage(m.id, m.msgType)
	if m.chunks != nil {
		_ = clone.addChunks(m.chunks)
	}
	m.clones = append(m.clones, clone)
	return clone
}

// Reader returns a io.Reader for the message
func (m *StreamMessage) Reader() io.Reader {
	return NewChunkReaderReader(m.ChunkReader())
}

// ChunkReader returns a ChunkReader for the message
func (m *StreamMessage) ChunkReader() ChunkReader {
	return &streamMessageChunkReader{m}
}
