package envelope

import (
	"io"

	"xbus.io/go-xbus/v4/api"
)

// NewMessageWriter returns a Message and a writer to write its content
// The Close method must be called to end the stream
func NewMessageWriter(msgType string) (MessageWriter, error) {
	id, err := api.UUIDV4()
	if err != nil {
		return nil, err
	}
	msgW := NewMessageWriterWithID(id, msgType)
	return msgW, nil
}

// NewMessageWriterWithID returns a Message and a writer to write its content
// The Close method must be called to end the stream
func NewMessageWriterWithID(id api.UUID, msgType string) MessageWriter {
	msg := NewStreamMessage(id, msgType)
	return newStreamMessageWriter(msg)
}

func MustNewMessageWriter(msgType string) MessageWriter {
	w, err := NewMessageWriter(msgType)
	if err != nil {
		panic(err)
	}
	return w
}

// newStreamMessageWriter implements io.WriteCloser and auto-split in chunks
func newStreamMessageWriter(msg *StreamMessage) *streamMessageWriter {
	return &streamMessageWriter{
		StreamMessage: msg,
		chunkWriter:   newChunkWriter(msg.addChunks),
	}
}

type streamMessageWriter struct {
	*StreamMessage
	*chunkWriter
}

func newChunkWriter(addChunks func([][]byte) error) *chunkWriter {
	return &chunkWriter{
		addChunks: addChunks,
	}
}

type chunkWriter struct {
	addChunks func([][]byte) error
	buf       []byte
	started   bool
}

func (w *chunkWriter) WriteChunk(chunk []byte) error {
	var chunks = make([][]byte, 0, 2)
	if len(w.buf) != 0 {
		chunks = append(chunks, w.buf)
		w.buf = w.buf[:0]
	}
	chunks = append(chunks, chunk)
	return w.addChunks(chunks)
}

func (w *chunkWriter) WriteStringChunk(chunk string) error {
	return w.WriteChunk([]byte(chunk))
}

func (w *chunkWriter) Write(data []byte) (int, error) {
	var datasize = len(data)
	if !w.started && len(w.buf)+len(data) < ChunkSplitThreadhold {
		w.buf = append(w.buf, data...)
		return datasize, nil
	}

	w.started = true

	var chunks [][]byte

	w.buf = append(w.buf, data[:]...)

	for len(w.buf) >= DefaultChunkSize {
		var (
			chunk []byte
			n     int
		)
		if len(w.buf) == DefaultChunkSize {
			chunk = w.buf
			n = len(w.buf)
		} else {
			chunk = w.buf[:DefaultChunkSize]
			n = DefaultChunkSize
		}
		w.buf = w.buf[n:]
		chunks = append(chunks, chunk)
	}

	if len(chunks) != 0 {
		if err := w.addChunks(chunks); err != nil {
			return 0, err
		}
	}
	return datasize, nil
}

func (w *chunkWriter) Close() error {
	if len(w.buf) != 0 {
		if err := w.addChunks([][]byte{w.buf}); err != nil {
			return err
		}
	}
	return w.addChunks(nil)
}

// DEPRECATED (to be removed soon)

// ChunkWriter writes chunks of message. The chunk sizes are kept and the exact
// same chunks can be read using a ChunkReader when reading the message.
// A single chunk should be small, and if too big will make the resulting
// message impossible to send on the bus. The recommanded maximum size is
// approximatively 768K.
type ChunkWriter interface {
	io.Closer
	Write([]byte) error
	WriteString(string) error
}

// NewMessageChunkWriter returns a Message and a chunk writer to write its content.
// The check writer gives controls on the chunk sizes, and will refuse chunks
// that are too big
// The Close method must be called to end the stream
func NewMessageChunkWriter(msgType string) (Message, ChunkWriter, error) {
	id, err := api.UUIDV4()
	if err != nil {
		return nil, nil, err
	}
	m, w := NewMessageChunkWriterWithID(id, msgType, 0)
	return m, w, nil
}

// MustNewMessageChunkWriter returns a Message and a chunk writer to write its content.
// It panics if a UUID cannot be generated
// The check writer gives controls on the chunk sizes, and will refuse chunks
// that are too big
// The Close method must be called to end the stream
func MustNewMessageChunkWriter(msgType string) (Message, ChunkWriter) {
	m, w, err := NewMessageChunkWriter(msgType)
	if err != nil {
		panic(err)
	}
	return m, w
}

// NewMessageChunkWriterWithID returns a Message and a chunk writer to write its content.
// If checksum is non-zero, it will be checked when the chunk writer is closed.
// The check writer gives controls on the chunk sizes, and will refuse chunks
// that are too big
// The Close method must be called to end the stream
func NewMessageChunkWriterWithID(
	id api.UUID, msgType string, checksum api.Checksum,
) (Message, ChunkWriter) {
	m := NewStreamMessage(id, msgType)
	if checksum != 0 {
		m.SetChecksum(checksum)
	}
	return m, NewStreamMessageChunkWriter(m.addChunks)
}

// NewStreamMessageChunkWriter implements ChunkWriter and output a chunk for reach call to Write
func NewStreamMessageChunkWriter(addChunks func([][]byte) error) ChunkWriter {
	return &streamMessageChunkWriter{addChunks}
}

type streamMessageChunkWriter struct {
	addChunks func([][]byte) error
}

func (w *streamMessageChunkWriter) Write(data []byte) error {
	var chunk = make([]byte, len(data))
	copy(chunk, data)
	return w.addChunks([][]byte{chunk})
}

func (w *streamMessageChunkWriter) WriteString(data string) error {
	return w.addChunks([][]byte{[]byte(data)})
}

func (w *streamMessageChunkWriter) Close() error {
	return w.addChunks(nil)
}
