package testutils

import (
	"context"
	"fmt"
	"sync"
	"testing"
	"time"

	nrpc "github.com/nats-rpc/nrpc"
	"github.com/stretchr/testify/assert"
	"google.golang.org/protobuf/types/known/timestamppb"
	xbus "xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/envelope"
)

const (
	// DefaultNodeID used in the ProcessingContext
	DefaultNodeID = "node"

	// DefaultActorName of the test actor
	DefaultActorName = "actor"
)

var (
	// DefaultProcessID used in the ProcessingContext
	DefaultProcessID = api.MustUUIDFromString("af75efb0-1f0d-11e9-9513-2329a2a94eaf")
	// DefaultActorID of the test actor
	DefaultActorID = api.MustUUIDFromString("7fe4f2b8-1f0e-11e9-8661-ebbdab78f251")
)

// ActorServiceTester helps testing a ActorProcessService service
type ActorServiceTester struct {
	serviceName string
	factory     xbus.NewActorServiceFunc
	Apc         xbus.ActorProcessingContext
	sendChannel chan *api.ActorProcessingState

	Outputs map[string]*xbus.EnvelopeReceiver
}

// NewActorServiceTester creates a ActorServiceTester
func NewActorServiceTester(
	t *testing.T,
	serviceName string,
	factory func(actor *xbus.Actor, settings xbus.Settings) (xbus.ActorService, error),
) *ActorServiceTester {
	logger := GetTestLogger(t)
	var tester = ActorServiceTester{
		serviceName: serviceName,
		factory:     factory,
		Apc: xbus.ActorProcessingContext{
			Context: context.Background(),
			Request: &api.ActorProcessRequest{
				Context: &api.ProcessingContext{
					ProcessID: DefaultProcessID.ToBytes(),
					NodeID:    DefaultNodeID,
				},
			},
			Actor: &xbus.Actor{
				ID:   DefaultActorID,
				Name: DefaultActorName,
			},
			Log: xbus.NewLogger(
				logger.With().
					Str("actor", DefaultActorName).
					Str("actor_id", DefaultActorID.String()).
					Logger(),
			),
		},
		Outputs: make(map[string]*xbus.EnvelopeReceiver),
	}
	xbus.ApcTestSetSend(
		&tester.Apc,
		func(s *api.ActorProcessingState) { tester.sendChannel <- s },
	)
	xbus.ApcTestSetCloseOutput(
		&tester.Apc,
		func(
			context *api.ProcessingContext, output string,
		) (*api.EnvelopeAck, error) {
			return &api.EnvelopeAck{
				Status: api.EnvelopeAck_ACCEPTED,
			}, tester.getOutputReceiver(output).Finalize()
		},
	)
	xbus.ApcTestSetSendFragment(
		&tester.Apc,
		func(
			ctx *api.ProcessingContext, output string, env *api.Envelope,
		) (api.EnvelopeAck_ReceptionStatus, error) {
			receiver := tester.getOutputReceiver(output)
			assert.NoError(t, receiver.AddFragment(env))
			return receiver.Status(), receiver.LastError()
		},
	)
	xbus.ApcTestSetProcessingEnd(
		&tester.Apc,
		func(
			context *api.ProcessingContext,
			status api.ActorProcessingState_Status,
			messages []*api.LogMessage,
		) error {
			tester.sendChannel <- &api.ActorProcessingState{
				Context:  context,
				Status:   status,
				Messages: messages,
			}
			return nil
		},
	)
	xbus.ApcTestSetReceiveEnvelope(&tester.Apc,
		func(
			ctx context.Context, clientKind, clientID string,
			envelopeID api.UUID,
			head *api.Envelope, position *api.EnvelopePosition,
		) (*envelope.BuilderFromFragment, error) {
			builder := envelope.NewFromFragments(envelopeID)

			if head != nil {
				if err := builder.Add(head); err != nil {
					return nil, err
				}

				if position == nil && !builder.Complete {
					assert.NoError(t, builder.Cancel(fmt.Errorf("No position passed, cannot retrieve the envelope remainings")))
					return nil, fmt.Errorf("No position passed, cannot retrieve the envelope remainings")
				}
			} // else {
			// p := api.EnvelopePositionStart(envelopeID)
			// position = &p
			//}
			if !builder.Complete {
				return nil, fmt.Errorf("streamed envelope are not yet supported in testutils")
			}
			return builder, nil
		})
	return &tester
}

// SetSettings sets the service settings in the Actor definition
func (t *ActorServiceTester) SetSettings(settings map[string]interface{}) {
	t.Apc.Actor.Options.Service.Settings = settings
}

// SetInputEnvelope associate an initial envelope fragment to an input
// If needed, further fragments can be fed during the RunProcess() with FeedInput
func (t *ActorServiceTester) SetInputEnvelope(name string, fragment *api.Envelope) {
	t.Apc.Request.Inputs = append(t.Apc.Request.Inputs, &api.ActorProcessRequest_Input{
		Name:     name,
		Envelope: fragment,
		Position: &api.EnvelopePosition{},
	})
}

// CloseInput marks an input as closed
func (t *ActorServiceTester) CloseInput(name string) {
	t.Apc.Request.Inputs = append(t.Apc.Request.Inputs, &api.ActorProcessRequest_Input{
		Name:  name,
		Close: true,
	})
}

// FeedInput adds a fragment to an input stream. Must be called _during_ Run to
// avoid being locked
func (t *ActorServiceTester) FeedInput(name string, fragment *api.Envelope) {
	panic("Not yet implemented")
}

// RunProcess runs Process() with the inputs if provided and store the outputs for further inspection
func (t *ActorServiceTester) RunProcess() (*api.ActorProcessingState, error) {
	var lastState *api.ActorProcessingState

	var (
		baseService xbus.ActorService
		err         error
	)
	if t.factory != nil {
		baseService, err = t.factory(t.Apc.Actor, t.Apc.Actor.Options.Service.Settings)
	} else {
		baseService, err = xbus.MakeActorService(t.serviceName, t.Apc.Actor)
	}
	if err != nil {
		return nil, err
	}
	service, ok := baseService.(xbus.ActorProcessService)
	if !ok {
		return nil, fmt.Errorf("Factory returned a non ActorProcessService")
	}
	if service == nil {
		return nil, fmt.Errorf("Factory returned nil")
	}
	t.sendChannel = make(chan *api.ActorProcessingState, 1)

	wg := sync.WaitGroup{}
	wg.Add(1)
	go func() {
		for state := range t.sendChannel {
			lastState = state
		}
		wg.Done()
	}()

	err = service.Process(&t.Apc)

	close(t.sendChannel)
	wg.Wait()

	if err != nil {
		if t.Apc.IsDone() {
			return lastState, fmt.Errorf("Process() returned an error but is marked as done")
		}
		if t.Apc.IsDetached() {
			return lastState, fmt.Errorf("Process() returned an error but is marked as detached")
		}
		if nErr, ok := err.(*nrpc.Error); ok && nErr.Type == nrpc.Error_SERVERTOOBUSY {
			return nil, nErr
		}
		if lastState == nil {
			lastState = &api.ActorProcessingState{
				Status: api.ActorProcessingState_ERROR,
				Messages: []*api.LogMessage{{
					Time:  timestamppb.Now(),
					Level: api.LogLevel_ERROR,
					Text:  err.Error(),
				}},
			}
		} else {
			lastState.Status = api.ActorProcessingState_ERROR
		}
	} else if lastState == nil {
		if !t.Apc.IsDetached() {
			lastState = &api.ActorProcessingState{
				Status: api.ActorProcessingState_SUCCESS,
			}
		}
	}

	return lastState, nil
}

func (t *ActorServiceTester) getOutputReceiver(name string) *xbus.EnvelopeReceiver {
	receiver, ok := t.Outputs[name]
	if !ok {
		receiver = xbus.NewEnvelopeReceiverEmpty()
		t.Outputs[name] = receiver
	}
	return receiver
}

// GetOutputEnvelope returns the complete envelope send on a given output
// Run must be over
func (t *ActorServiceTester) GetOutputEnvelope(name string, timeout time.Duration) (*api.Envelope, error) {
	return t.getOutputReceiver(name).Complete(timeout)
}

// GetOutputReceiver returns an envelope receiver for envelope send on a given
// output
func (t *ActorServiceTester) GetOutputReceiver(name string) *xbus.EnvelopeReceiver {
	return t.getOutputReceiver(name)
}
