package utils

import (
	"time"

	jsoniter "github.com/json-iterator/go"
)

func NewJSONStreamObject(stream *jsoniter.Stream) *JSONStreamObject {
	return &JSONStreamObject{
		stream:  stream,
		isEmpty: true,
	}
}

type JSONStreamObject struct {
	stream  *jsoniter.Stream
	isEmpty bool
}

func (so *JSONStreamObject) End() {
	if so.isEmpty {
		so.stream.WriteEmptyObject()
	} else {
		so.stream.WriteObjectEnd()
	}
}

func (so *JSONStreamObject) Field(name string) {
	if so.isEmpty {
		so.isEmpty = false
		so.stream.WriteObjectStart()
	} else {
		so.stream.WriteMore()
	}
	so.stream.WriteObjectField(name)
}

func (so *JSONStreamObject) Nil(name string) {
	so.Field(name)
	so.stream.WriteNil()
}

func (so *JSONStreamObject) True(name string) {
	so.Field(name)
	so.stream.WriteTrue()
}

func (so *JSONStreamObject) False(name string) {
	so.Field(name)
	so.stream.WriteFalse()
}

func (so *JSONStreamObject) Bool(name string, value bool) {
	so.Field(name)
	so.stream.WriteBool(value)
}

func (so *JSONStreamObject) OEBool(name string, value bool) {
	if value {
		so.True(name)
	}
}

func (so *JSONStreamObject) Object(name string) *JSONStreamObject {
	so.Field(name)
	return NewJSONStreamObject(so.stream)
}

func (so *JSONStreamObject) Array(name string) *JSONStreamArray {
	so.Field(name)
	return NewJSONStreamArray(so.stream)
}

func (so *JSONStreamObject) StringWithHTMLEscaped(name string, value string) {
	so.Field(name)
	so.stream.WriteStringWithHTMLEscaped(value)
}

func (so *JSONStreamObject) OEStringWithHTMLEscaped(name string, value string) {
	if value != "" {
		so.StringWithHTMLEscaped(name, value)
	}
}

func (so *JSONStreamObject) String(name string, value string) {
	so.Field(name)
	so.stream.WriteString(value)
}

func (so *JSONStreamObject) OEString(name string, value string) {
	if value != "" {
		so.String(name, value)
	}
}

func (so *JSONStreamObject) StringList(name string, values []string) {
	a := so.Array(name)
	for _, value := range values {
		a.String(value)
	}
	a.End()
}

func (so *JSONStreamObject) OEStringList(name string, values []string) {
	if len(values) != 0 {
		so.StringList(name, values)
	}
}

func (so *JSONStreamObject) Uint8(name string, value uint8) {
	so.Field(name)
	so.stream.WriteUint8(value)
}

func (so *JSONStreamObject) OEUint8(name string, value uint8) {
	if value != 0 {
		so.Uint8(name, value)
	}
}

func (so *JSONStreamObject) Int8(name string, value int8) {
	so.Field(name)
	so.stream.WriteInt8(value)
}

func (so *JSONStreamObject) OEInt8(name string, value int8) {
	if value != 0 {
		so.Int8(name, value)
	}
}

func (so *JSONStreamObject) Uint16(name string, value uint16) {
	so.Field(name)
	so.stream.WriteUint16(value)
}

func (so *JSONStreamObject) OEUint16(name string, value uint16) {
	if value != 0 {
		so.Uint16(name, value)
	}
}

func (so *JSONStreamObject) Int16(name string, value int16) {
	so.Field(name)
	so.stream.WriteInt16(value)
}

func (so *JSONStreamObject) OEInt16(name string, value int16) {
	if value != 0 {
		so.Int16(name, value)
	}
}

func (so *JSONStreamObject) Uint32(name string, value uint32) {
	so.Field(name)
	so.stream.WriteUint32(value)
}

func (so *JSONStreamObject) OEUint32(name string, value uint32) {
	if value != 0 {
		so.Uint32(name, value)
	}
}

func (so *JSONStreamObject) Int32(name string, value int32) {
	so.Field(name)
	so.stream.WriteInt32(value)
}

func (so *JSONStreamObject) OEInt32(name string, value int32) {
	if value != 0 {
		so.Int32(name, value)
	}
}

func (so *JSONStreamObject) Uint64(name string, value uint64) {
	so.Field(name)
	so.stream.WriteUint64(value)
}

func (so *JSONStreamObject) OEUint64(name string, value uint64) {
	if value != 0 {
		so.Uint64(name, value)
	}
}

func (so *JSONStreamObject) Int64(name string, value int64) {
	so.Field(name)
	so.stream.WriteInt64(value)
}

func (so *JSONStreamObject) OEInt64(name string, value int64) {
	if value != 0 {
		so.Int64(name, value)
	}
}

func (so *JSONStreamObject) Uint(name string, value uint) {
	so.Field(name)
	so.stream.WriteUint(value)
}

func (so *JSONStreamObject) OEUint(name string, value uint) {
	if value != 0 {
		so.Uint(name, value)
	}
}

func (so *JSONStreamObject) Int(name string, value int) {
	so.Field(name)
	so.stream.WriteInt(value)
}

func (so *JSONStreamObject) OEInt(name string, value int) {
	if value != 0 {
		so.Int(name, value)
	}
}

func (so *JSONStreamObject) Float32(name string, value float32) {
	so.Field(name)
	so.stream.WriteFloat32(value)
}

func (so *JSONStreamObject) OEFloat32(name string, value float32) {
	if value != 0 {
		so.Float32(name, value)
	}
}

func (so *JSONStreamObject) Float32Lossy(name string, value float32) {
	so.Field(name)
	so.stream.WriteFloat32Lossy(value)
}

func (so *JSONStreamObject) OEFloat32Lossy(name string, value float32) {
	if value != 0 {
		so.Float32Lossy(name, value)
	}
}

func (so *JSONStreamObject) Float64(name string, value float64) {
	so.Field(name)
	so.stream.WriteFloat64(value)
}

func (so *JSONStreamObject) OEFloat64(name string, value float64) {
	if value != 0 {
		so.Float64(name, value)
	}
}

func (so *JSONStreamObject) Float64Lossy(name string, value float64) {
	so.Field(name)
	so.stream.WriteFloat64Lossy(value)
}

func (so *JSONStreamObject) OEFloat64Lossy(name string, value float64) {
	if value != 0 {
		so.Float64Lossy(name, value)
	}
}

func (so *JSONStreamObject) Time(name string, value time.Time) {
	so.Field(name)
	so.stream.WriteString(value.Format(time.RFC3339))
}

func (so *JSONStreamObject) TimeOE(name string, value time.Time) {
	if !value.IsZero() {
		so.Time(name, value)
	}
}

func (so *JSONStreamObject) Val(name string, value any) {
	so.Field(name)
	if m, ok := value.(JSONStreamMarshaler); ok {
		m.MarshalJSONStream(so.stream)
	} else {
		so.stream.WriteVal(value)
	}
}
